"""A collection of monsters. Also useful for building a list of wild
shape forms."""


from .stats import Ability


class Monster():
    """A monster that may be encountered when adventuring."""
    name = "Generic Monster"
    description = ""
    challenge_rating = 0
    armor_class = 0
    skills = "Perception +3, Stealth +4"
    senses = ""
    languages = ""
    strength = Ability()
    dexterity = Ability()
    constitution = Ability()
    intelligence = Ability()
    wisdom = Ability()
    charisma = Ability()
    speed = 30
    swim_speed = 0
    fly_speed = 0
    hp_max = 10
    hit_dice = '1d6'
    
    @property
    def is_beast(self):
        is_beast = 'beast' in self.description.lower()
        return is_beast


class Ankylosaurus(Monster):
    """Thick armor plating covers the body of the plant-eating dinosaur
    ankylosaurus, which defends itself against predators with a
    knobbed tail that delivers a devastating strike.
    
    **Tail:** *Melee Weapon Attack:* +7 to hit, reach 10 ft., one
    target. *Hit:* 18 (4d6+4) bludgeoning damage. If the target is a
    creature, it must succeed on a DC 14 Strength saving throw or be
    knocked prone.
    
    """
    name = "Ankylosaurus"
    description = "Huge beast, unaligned"
    challenge_rating = 3
    armor_class = 15
    skills = ""
    senses = "Passive perception 11"
    strength = Ability(19)
    dexterity = Ability(11)
    constitution = Ability(15)
    intelligence = Ability(2)
    wisdom = Ability(12)
    charisma = Ability(5)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    hp_max = 68
    hit_dice = '8d12+16'


class Ape(Monster):
    """**Multiattack:** The ape makes two fist attacks.
    
    **Fist:** *Melee Weapon Attack:* +5 to hit, reach 5 ft., one
    target. *Hit:* 6 (1d6+3) bludgeoning damage.
    
    **Rock:** *Ranged Weapon Attack:* +5 to hit, range 25/50 ft., one
    target. *Hit:* 6 (1d6+3) bludgeoning damage.
    
    """
    name = "Ape"
    description = "Medium beast, unaligned"
    challenge_rating = 1 / 2
    armor_class = 12
    skills = "Athletics +5, Perception +3"
    senses = "Passive perception 13"
    strength = Ability(16)
    dexterity = Ability(14)
    constitution = Ability(14)
    intelligence = Ability(6)
    wisdom = Ability(12)
    charisma = Ability(7)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    hp_max = 19
    hit_dice = '3d8+6'


class Crocodile(Monster):
    """**Hold Breath:** The crocodile can hold its breath for 15 minutes.
    
    **Bite:** *Melee Weapon Attack:* +4 to hit, reach 5 ft., one
      creature. *Hit:* 7 (1d10+2) piercing damage, and the target is
      Grappled (escape DC 12). Until this grapple ends, the target is
      Restrained, and the crocodile can't bite another target.
    
    """
    name = "Crocodile"
    description = "Large beast, unaligned"
    challenge_rating = 1/2
    armor_class = 12
    skills = "Stealth +2"
    senses = "Passive perception 10"
    strength = Ability(15)
    dexterity = Ability(10)
    constitution = Ability(13)
    intelligence = Ability(2)
    wisdom = Ability(10)
    charisma = Ability(5)
    speed = 30
    swim_speed = 30
    fly_speed = 0
    hp_max = 19
    hit_dice = '3d10+3'

class GiantEagle(Monster):
    """A giant eagle is a noble creature that speaks its own language and
    understands Speech in the Common tongue. A mated pair of giant
    eagles typically has up to four eggs or young in their nest (treat
    the young as normal eagles).
    
    **Keen Sight:** The eagle has advantage on Wisdom (Perception)
      checks that rely on sight.
    
    **Multiattack:** The eagle makes two attacks: one with its beak
      and one with its talons.
    
    **Beak:** *Melee Weapon Attack:* +5 to hit, reach 5 ft., one
      target. *Hit:* 6 (1d6 + 3) piercing damage.
    
    **Talons:** *Melee Weapon Attack:* +5 to hit, reach 5 ft., one
      target. *Hit:* 10 (2d6 + 3) slashing damage.
    
    """
    name = "Giant eagle"
    description = "Large beast, neutral good"
    challenge_rating = 1
    armor_class = 13
    skills = "Perception +4"
    senses = "Passive perception 14"
    languages = "Giant Eagle, understands common and Auran but can't speak."
    strength = Ability(16)
    dexterity = Ability(17)
    constitution = Ability(13)
    intelligence = Ability(8)
    wisdom = Ability(14)
    charisma = Ability(10)
    speed = 10
    swim_speed = 0
    fly_speed = 80
    hp_max = 26
    hit_dice = '4d10+4'


class Spider(Monster):
    """**Spider Climb:** The spider can climb difficult surfaces,
    including upside down on ceilings, without needing to make an
    ability check.
    
    **Web Sense:** While in contact with a web, the spider knows the
    exact location of any other creature in contact with the same web.
    
    **Web Walker:** The spider ignores Movement restrictions caused by
    webbing.
    
    **Bite:** Melee Weapon Attack: +4 to hit, reach 5 ft., one
    creature. Hit: 1 piercing damage, and the target must succeed on a
    DC 9 Constitution saving throw or take 2 (1d4) poison damage.
    
    """
    name = "Spider"
    description = "Tiny beast, unaligned"
    challenge_rating = 0
    armor_class = 12
    skills = "Stealth +4"
    senses = "Darkvision 30 Ft., passive Perception 12"
    strength = Ability(2)
    dexterity = Ability(14)
    constitution = Ability(8)
    intelligence = Ability(1)
    wisdom = Ability(10)
    charisma = Ability(2)
    speed = 20
    hp_max = 1
    hit_dice = '1d4-1'


class SwarmOfRats(Monster):
    """**Keen Smell:** The swarm has advantage on Wisdom (Perception)
    checks that rely on smell.
    
    **Swarm:** The swarm can occupy another creature's space and vice
      versa, and the swarm can move through any opening large enough
      for a Tiny rat. The swarm can't regain Hit Points or gain
      Temporary Hit Points.
    
    **Bites:** Melee Weapon Attack: +2 to hit, reach 0 ft., one target
      in the swarm's space. Hit: 7 (2d6) piercing damage, or 3 (1d6)
      piercing damage if the swarm has half of its Hit Points or
      fewer.
    
    """
    name = "Swarm of Rats"
    description = "Medium swarm of tiny beasts, unaligned"
    challenge_rating = 1/4
    armor_class = 10
    skills = ""
    senses = "Darkvision 30 Ft., passive Perception 10"
    damage_resistance = "Bludgeoning, Piercing, Slashing"
    condition_immunities = ("Charmed, Frightened, Grappled, Paralyzed, "
                            "Petrified, Prone, Restrained, Stunned")
    strength = Ability(9)
    dexterity = Ability(11)
    constitution = Ability(9)
    intelligence = Ability(2)
    wisdom = Ability(10)
    charisma = Ability(3)
    speed = 30
    hp_max = 24
    hit_dice = '7d8-7'


class Rat(Monster):
    """**Keen Smell:** The rat has advantage on Wisdom (Perception) checks
    that rely on smell.
    
    **Bite:** Melee Weapon Attack: +0 to hit, reach 5 ft., one
    target. Hit: 1 piercing damage.
    
    """
    name = "Rat"
    description = "Tiny beast, unaligned"
    challenge_rating = 0
    armor_class = 10
    skills = ""
    senses = "Darkvision 30 Ft., passive Perception 10"
    strength = Ability(2)
    dexterity = Ability(11)
    constitution = Ability(9)
    intelligence = Ability(2)
    wisdom = Ability(10)
    charisma = Ability(4)
    speed = 20
    hp_max = 1
    hit_dice = '1d4-2'


class Wolf(Monster):
    """**Keen Hearing and Smell.** The wolf has advantage on Wisdom
    (Perception) checks that rely on hearing or smell.
    
    **Pack Tactics.** The wolf has advantage on an attack roll against a
    creature if at least one of the wolf's allies is within 5 ft. of
    the creature and the ally isn't incapacitated.  Actions
    
    **Bite.** *Melee Weapon Attack:* +4 to hit, reach 5 ft., one
    target. *Hit:* (2d4+2) piercing damage. If the target is a
    creature, it must succeed on a DC 11 Strength saving throw or be
    knocked prone
    
    """
    name = "Wolf"
    description = "Medium beast, unaligned"
    challenge_rating = 1/4
    armor_class = 13
    skills = "Perception +3, Stealth +4"
    senses = "Passive Perception 13"
    strength = Ability(12)
    dexterity = Ability(15)
    constitution = Ability(12)
    intelligence = Ability(6)
    wisdom = Ability(12)
    charisma = Ability(6)
    speed = 40
    hp_max = 11
    hit_dice = '2d8+2'



from . import (weapons, spells)
from . import features as feats
from collections import defaultdict


class Race():
    name = "Unknown"
    size = "medium"
    speed = 30
    owner = None
    languages = ('Common', )
    proficiencies_text = tuple()
    weapon_proficiences = tuple()
    skill_proficiencies = ()
    skill_choices = ()
    num_skill_choices = 0
    features = tuple()
    features_by_level = defaultdict(list)
    strength_bonus = 0
    dexterity_bonus = 0
    constitution_bonus = 0
    intelligence_bonus = 0
    wisdom_bonus = 0
    charisma_bonus = 0
    hit_point_bonus = 0
    spells_known = ()

    def __init__(self, owner=None):
        self.owner = owner
        cls = type(self)
        # Instantiate the features
        self.features = tuple([f(owner=self.owner) for f in cls.features])
        self.features_by_level = defaultdict(list)
        for i in range(1, 21):
            self.features_by_level[i] = [f(owner=self.owner)for f in
                                         cls.features_by_level[i]]
        self.spells_known = [S() for S in cls.spells_known]

    @property
    def spells_prepared(self):
        return self.spells_known
    
    def __str__(self):
        return self.name
    
    def __repr__(self):
        return "\"{:s}\"".format(self.name)


# Dwarves
class _Dwarf(Race):
    name = "Dwarf"
    size = "medium"
    speed = 25
    languages = ("Common", "Dwarvish")
    constitution_bonus = 2
    proficiencies_text = ('battleaxes', 'handaxes', 'throwing hammers',
                          'warhammers')
    weapon_proficiences = (weapons.Battleaxe, weapons.Handaxe,
                           weapons.ThrowingHammer, weapons.Warhammer)
    features = (feats.Darkvision, feats.DwarvenResilience, feats.Stonecunning)


class HillDwarf(_Dwarf):
    name = "Hill Dwarf"
    wisdom_bonus = 1
    hit_point_bonus = 1
    features = _Dwarf.features + (feats.DwarvenToughness,)


class MountainDwarf(_Dwarf):
    name = "Mountain Dwarf"
    strength_bonus = 2


# Elves
class _Elf(Race):
    name = "Elf"
    size = "medium"
    speed = 30
    dexterity_bonus = 2
    skill_proficiencies = ('perception',)
    languages = ('Common', 'Elvish')
    features = (feats.Darkvision, feats.FeyAncestry, feats.Trance)


class HighElf(_Elf):
    name = "High Elf"
    weapon_proficiencies = (weapons.Longsword, weapons.Shortsword,
                            weapons.Shortbow, weapons.Longbow)
    proficiencies_text = ('longswords', 'shortswords', 'shortbows', 'longbows')
    intelligence_bonus = 1
    languages = ('Common', 'Elvish', '[choose one]')
    features = _Elf.features + (feats.ElfCantrip,)


class WoodElf(_Elf):
    name = "Wood Elf"
    weapon_proficiencies = (weapons.Longsword, weapons.Shortsword,
                            weapons.Shortbow, weapons.Longbow)
    proficiencies_text = ('longswords', 'shortswords', 'shortbows', 'longbows')
    wisdom_bonus = 1
    speed = 35
    features = _Elf.features + (feats.MaskOfTheWild,)


class DarkElf(_Elf):
    name = "Dark Elf"
    weapon_proficiencies = (weapons.Rapier, weapons.Shortsword,
                            weapons.HandCrossbow)
    proficiencies_text = ('rapiers', 'shortswords', 'hand crossbows')
    charisma_bonus = 1
    features = (feats.SuperiorDarkvision, feats.FeyAncestry, feats.Trance,
                feats.SunlightSensitivity, feats.DrowMagic)
    spells_known = (spells.DancingLights,)


# Halflings
class _Halfling(Race):
    name = "Halfling"
    size = "small"
    speed = 25
    dexterity_bonus = 2
    languages = ('Common', 'Halfling')
    features = (feats.Lucky, feats.Brave, feats.HalflingNimbleness)


class LightfootHalfling(_Halfling):
    name = "Lightfoot Halfling"
    charisma_bonus = 1
    features = _Halfling.features + (feats.NaturallyStealthy,)
    

class StoutHalfling(_Halfling):
    name = "Stout Halfling"
    constitution_bonus = 1
    features = _Halfling.features + (feats.StoutResilience,)


# Humans
class Human(Race):
    name = "Human"
    size = "medium"
    speed = 30
    strength_bonus = 1
    dexterity_bonus = 1
    constitution_bonus = 1
    intelligence_bonus = 1
    wisdom_bonus = 1
    charisma_bonus = 1
    languages = ("Common", '[choose one]')


# Dragonborn
class Dragonborn(Race):
    name = "Dragonborn"
    size = "medium"
    speed = 30
    strength_bonus = 2
    charisma_bonus = 1
    languages = ("Common", "Draconic")
    features = (feats.DraconicAncestry, feats.BreathWeapon,
                feats.DraconicResistance)


# Gnomes
class _Gnome(Race):
    name = "Gnome"
    size = "small"
    speed = 25
    intelligence_bonus = 2
    languages = ("Common", "Gnomish")
    features = (feats.Darkvision, feats.GnomeCunning)


class ForestGnome(_Gnome):
    name = "Forest Gnome"
    dexterity_bonus = 1
    features = _Gnome.features + (feats.NaturalIllusionist,
                                  feats.SpeakWithSmallBeasts)
    spells_known = (spells.MinorIllusion,)
    

class RockGnome(_Gnome):
    name = "Rock Gnome"
    constitution_bonus = 1
    features = _Gnome.features + (feats.ArtificersLore,
                                  feats.Tinker)


class DeepGnome(_Gnome):
    name = "Deep Gnome"
    dexterity_bonus = 1
    languages = ("Common", "Gnomish", "Undercommon")
    features = (feats.SuperiorDarkvision, feats.GnomeCunning,
                feats.StoneCamouflage)


# Half-elves
class HalfElf(Race):
    name = "Half-Elf"
    size = "medium"
    speed = 30
    charisma_bonus = 2
    skill_choices = ('acrobatics', 'animal handling', 'arcana',
                     'athletics', 'deception', 'history', 'insight',
                     'intimidation', 'investigation', 'medicine', 'nature',
                     'perception', 'performance', 'persuasion', 'religion',
                     'sleight of hand', 'stealth', 'survival')
    num_skill_choices = 2
    languages = ("Common", "Elvish", "[choose one]")
    features = (feats.Darkvision, feats.FeyAncestry)


# Half-Orcs
class HalfOrc(Race):
    name = "Half-Orc"
    size = "medium"
    speed = 30
    strength_bonus = 2
    constitution_bonus = 1
    skill_proficiencies = ('intimidation',)
    languages = ("Common", "Orc")
    features = (feats.Darkvision, feats.RelentlessEndurance,
                feats.SavageAttacks)


# Tielflings
class Tiefling(Race):
    name = "Tiefling"
    size = "medium"
    speed = 30
    intelligence_bonus = 1
    charisma_bonus = 2
    languages = ("Common", "Infernal")
    features = (feats.Darkvision, feats.HellishResistance,
                feats.InfernalLegacy)


# Aassimar
class _Aasimar(Race):
    name = 'Aasimar'
    size = 'medium'
    speed = 30
    charisma_bonus = 2
    languages = ("Common", "Celestial")
    features = (feats.Darkvision, feats.CelestialResistance,
                feats.HealingHands, feats.LightBearer)
    spells_known = (spells.Light,)

    
# Protector Aasimar
class ProtectorAasimar(_Aasimar):
    name = "Protector Aasimar"
    wisdom_bonus = 1
    features_by_level = defaultdict(list)
    features_by_level[3] += [feats.RadiantSoul]


# Fallen Aasimar
class ScourgeAasimar(_Aasimar):
    name = "Scourge Aasimar"
    constitution_bonus = 1
    features_by_level = defaultdict(list)
    features_by_level[3] += [feats.RadiantConsumption]


# Fallen Aasimar
class FallenAasimar(_Aasimar):
    name = "Fallen Aasimar"
    strength_bonus = 1
    features_by_level = defaultdict(list)
    features_by_level[3] += [feats.NecroticShroud]


# Firbolg
class Firbolg(Race):
    name = "Firbolg"
    size = "medium"
    speed = 30
    wisdom_bonus = 2
    strength_bonus = 1
    features = (feats.FirbolgMagic, feats.HiddenStep,
                feats.PowerfulBuild, feats.SpeechOfBeastAndLeaf)
    languages = ("Common", "Elvish", "Giant")


# Goliath
class Goliath(Race):
    name = "Goliath"
    size = "Medium"
    speed = 30
    skill_proficiencies = ("athletics",)
    languages = ("Common", "Giant")
    features = (feats.StonesEndurance, feats.PowerfulBuild,
                feats.MountainBorn)


# Lizardfolk
class Lizardfolk(Race):
    name = 'Lizardfolk'
    size = 'medium'
    speed = """30 (30 swim)"""
    constitution_bonus = 2
    wisdom_bonus = 1
    languages = ('Common', 'Draconic')
    weapon_proficiencies = (weapons.Bite,)
    proficiencies_text = ('bite',)
    features = (feats.CunningArtisan, feats.HoldBreath,
                feats.NaturalArmor, feats.HungryJaws)
    skill_choices = ('animal handling', 'nature', 'perception',
                     'stealth', 'survival')

    def __init__(self, owner=None):
        super().__init__(owner=owner)
        self.owner.wield_weapon("bite")


# Kenku
class Kenku(Race):
    name = 'Kenku'
    size = 'medium'
    speed = 30
    dexterity_bonus = 2
    wisdom_bonus = 1
    languages = ('Common', 'Auran')
    skill_choices = ('acrobatics', 'deception', 'stealth',
                     'sleight of hand')
    num_skill_choices = 2
    features = (feats.ExpertForgery, feats.Mimicry,)


# Tabaxi
class Tabaxi(Race):
    name = 'Tabaxi'
    size = 'medium'
    dexterity_bonus = 2
    charisma_bonus = 1
    speed = "30 (20 climb)"
    languages = ("Common", "[Choose One]")
    weapon_proficiencies = (weapons.Claws,)
    proficiences_text = ('Claws',)
    skill_proficiencies = ('perception', 'stealth')
    features = (feats.Darkvision, feats.FelineAgility,)

    def __init__(self, owner=None):
        super().__init__(owner=owner)
        self.owner.wield_weapon("claws")


# Triton
class Triton(Race):
    name = "Triton"
    size = "medium"
    strength_bonus = 1
    constitution_bonus = 1
    charisma_bonus = 1
    speed = "30 (30 swim)"
    features = (feats.Amphibious, feats.ControlAirAndWater,
                feats.EmissaryOfTheSea, feats.GuardiansOfTheDepths)
    languages = ("Common", "Primordial")
    spells_known = (spells.FogCloud,)


# Aarakocra
class Aarakocra(Race):
    name = 'Aarakocra'
    size = 'medium'
    speed = "25 (50 fly)"
    dexterity_bonus = 2
    wisdom_bonus = 1
    languages = ('Common', 'Aarakocra', 'Auran')
    weapon_proficiencies = (weapons.Talons,)
    proficiences_text = ('Talons',)

    def __init__(self, owner=None):
        super().__init__(owner=owner)
        self.owner.wield_weapon("talons")


# Genasi
class _Genasi(Race):
    name = "Genasi"
    constitution_bonus = 2
    size = 'medium'
    speed = 30
    languages = ("Common", 'Primoridal')


class AirGenasi(_Genasi):
    name = "Air Genasi"
    dexterity_bonus = 1
    features = (feats.UnendingBreath,
                feats.MingleWithTheWind)

    
class EarthGenasi(_Genasi):
    name = "Earth Genasi"
    strength_bonus = 1
    features = (feats.EarthWalk, feats.MergeWithStone)


class FireGenasi(_Genasi):
    name = "Fire Genasi"
    intelligence_bonus = 1
    features = (feats.Darkvision, feats.FireResistance,
                feats.ReachToTheBlaze)


class WaterGenasi(_Genasi):
    name = "Water Genasi"
    wisdom_bonus = 1
    speed = "30 (30 swim)"
    features = (feats.AcidResistance, feats.Amphibious,
                feats.CallToTheWave)


PHB_races = [HillDwarf, MountainDwarf, HighElf, WoodElf, DarkElf,
             LightfootHalfling, StoutHalfling, Human, Dragonborn,
             ForestGnome, RockGnome, HalfElf, HalfOrc, Tiefling]

VOLO_races = [ProtectorAasimar, ScourgeAasimar, FallenAasimar,
              Firbolg, Goliath, Lizardfolk, Kenku, Tabaxi, Triton]

EE_races = [Aarakocra, DeepGnome, AirGenasi, FireGenasi, EarthGenasi,
            WaterGenasi]

available_races = PHB_races + VOLO_races + EE_races

__all__ = tuple([r.name for r in available_races]) + (
    'available_races', 'PHB_races', 'VOLO_races', 'EE_races')



from .spells import Spell


class EarthTremor(Spell):
    """You cause a tremor in the ground in a 10-foot radius. Each creature other than 
    you in that area must make a Dexterity saving throw. On a failed save, a 
    creature takes 1d6 bludgeoning damage and is knocked prone. If the ground in 
    that area is loose earth or stone, it becomes difficult terrain until cleared. 
    At Higher Levels. When you cast this spell using a spell slot of 2nd level or 
    higher, the damage increases by 1d6 for each slot level above 1st.
    """
    name = "Earth Tremor"
    level = 1
    casting_time = "1 action"
    casting_range = "Self (10-foot radius)"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Bard', 'Druid', 'Sorcerer', 'Wizard')


class Earthbind(Spell):
    """Choose one creature you can see within range. Yellow strips of magical energy 
    loop around the creature. The target must succeed on a Strength saving throw or 
    its flying speed (if any) is reduced to 0 feet for the spell’s duration. An 
    airborne creature affected by this spell descends at 60 feet per round until it 
    reaches the ground or the spell ends.
    """
    name = "Earthbind"
    level = 2
    casting_time = "1 action"
    casting_range = "300 feet"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Sorcerer', 'Warlock', 'Wizard')


class Earthquake(Spell):
    """You create a seismic disturbance at a point on the ground that you can see 
    within range. 
    For the duration, an intense tremor rips through the ground in a 
    100-foot- radius circle centered on that point and shakes creatures and 
    structures in contact with the ground in that area. 
    
    The ground in the area 
    becomes difficult terrain. Each creature on the ground that is concentrating 
    must make a Constitution saving throw. On a failed save, the creature’s 
    concentration is broken. 
    
    When you cast this spell and at the end of each turn 
    you spend concentrating on it, each creature on the ground in the area must make
     a Dexterity saving throw. On a failed save, the creature is knocked prone. 
    
    
    This spell can have additional effects depending on the terrain in the area, as 
    determined by the DM. 
    Fissures. 
    Fissures open throughout the spell’s area at 
    the start of your next turn after you cast the spell. A total of 1d6 such 
    fissures open in locations chosen by the DM. Each is 1d10 x 10 feet deep, 10 
    feet wide, and extends from one edge of the spell’s area to the opposite side. A
     creature standing on a spot where a fissure opens must succeed on a Dexterity 
    saving throw or fall in. A creature that successfully saves moves with the 
    fissure’s edge as it opens. 
    A fissure that opens beneath a structure causes it 
    to automatically collapse (see below). 
    
    Structures. 
    The tremor deals 50 
    bludgeoning damage to any structure in contact with the ground in the area when 
    you cast the spell and at the start of each of your turns until the spell ends. 
    If a structure drops to 0 hit points, it collapses and potentially damages 
    nearby creatures. A creature within half the distance of a structure’s height 
    must make a Dexterity saving throw. On a failed save, the creature takes 5d6 
    bludgeoning damage, is knocked prone, and is buried in the rubble, requiring a 
    DC 20 Strength (Athletics) check as an action to escape. The DM can adjust the 
    DC higher or lower, depending on the nature of the rubble. On a successful save,
     the creature takes half as much damage and doesn’t fall prone or become buried.
    """
    name = "Earthquake"
    level = 8
    casting_time = "1 action"
    casting_range = "500 feet"
    components = ('V', 'S', 'M')
    materials = """A pinch o f dirt, a piece o f rock, and a lump of clay"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric', 'Druid', 'Sorcerer')


class EldritchBlast(Spell):
    """A beam of crackling energy streaks toward a creature within range. Make a ranged
     spell attack against the target. On a hit, the target takes 1d10 force damage.
    
    
    At Higher Levels: The spell creates more than one beam when you reach higher 
    levels: 
    Two beams at 5th level
    Three beams at 11th level
    Four beams at 17th 
    level. 
    You can direct the beams at the same target or at different ones. Make 
    a separate attack roll for each beam.
    """
    name = "Eldritch Blast"
    level = 0
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Warlock',)


class ElementalBane(Spell):
    """Choose one creature you can see within range, and choose one of the following 
    damage types: acid, cold, fire, lightning, or thunder.
    The target must succeed 
    on a Constitution saving throw or be affected by the spell for its duration. The
     first time each turn the affected target takes damage of the chosen type, the 
    target takes an extra 2d6 damage of that type. Moreover, the target loses any 
    resistance to that damage type until the spell ends.
    
    At Higher Levels: When you
     cast this spell using a spell slot of 5th level or higher, you can target one 
    additional creature for each slot level above 4th. The creatures must be within 
    30 feet of each other when you target them.
    """
    name = "Elemental Bane"
    level = 4
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Wizard', 'Warlock')


class ElementalWeapon(Spell):
    """A nonmagical weapon you touch becomes a magic weapon. 
    Choose one of the 
    following damage types: acid, cold, fire, lightning, or thunder. For the 
    duration, the weapon has a +1 bonus to attack rolls and deals an extra 1d4 
    damage of the chosen type when it hits.
    
    At Higher Levels: When you cast this 
    spell using a spell slot of 5th or 6th level, the bonus to attack rolls 
    increases to +2 and the extra damage increases to 2d4. 
    When you use a spell 
    slot of 7th level or higher, the bonus increases to +3 and the extra damage 
    increases to 3d4.
    """
    name = "Elemental Weapon"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Paladin',)


class EnemiesAbound(Spell):
    """You reach into the mind of one creature you can see and force it to make an 
    Intelligence saving throw. A creature automatically succeeds if it is immune to 
    being frightened. On a failed save, the target loses the ability to distinguish 
    friend from foe, regarding all creatures it can see as enemies until the spell 
    ends. Each time the target takes damage, it can repeat the saving throw, ending 
    the effect on itself on a success. Whenever the affected creature chooses 
    another creature as a target, it must choose the target at random from among the
     creatures it can see within range of the attack, spell, or other ability it’s 
    using. If an enemy provokes an opportunity attack from the affected creature, 
    the creature must make that attack if it is able to.
    """
    name = "Enemies Abound"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class Enervation(Spell):
    """A tendril of inky darkness reaches out from you, touching a creature you can see
     within range to drain life from it. The target must make a Dexterity saving 
    throw. On a successful save, the target takes 2d8 necrotic damage, and the spell
     ends. On a failed save, the target takes 4d8 necrotic damage, and until the 
    spell ends, you can use your action on each of your turns to automatically deal 
    4d8 necrotic damage to the target. The spell ends ifyou use your action to do 
    anything else, if the target is ever outside the spell’s range, or if the target
     has total cover from you. Whenever the spell deals damage to a target, you 
    regain hit points equal to half the amount of necrotic damage the target takes.
    
    
    At Higher Levels: When you cast this spell using a spell slot of 6th level or 
    higher, the damage increases by 1d8 for each slot level above 5th.
    """
    name = "Enervation"
    level = 5
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class EnhanceAbility(Spell):
    """You touch a creature and bestow upon it a magical enhancement. Choose one of the
     following effects: the target gains the effect until the spell ends.
    - Bear’s 
    Endurance. The target has advantage on Constitution checks. It also gains 2d6 
    temporary hit points, which are lost when the spell ends.
    - Bull’s Strength. The
     target has advantage on Strength checks, and his or her carrying capacity 
    doubles.
    - Cat’s Grace. The target has advantage on Dexterity checks. It also 
    doesn’t take damage from falling 20 feet or less if it isn’t incapacitated.
    - 
    Eagle’s Splendor. The target has advantage on Charisma checks.
    - Fox’s Cunning. 
    The target thas advantage on Intelligence checks.
    - Owl’s Wisdom. The target has
     advantage on Wisdom checks.
    
    At Higher Levels: When you cast this spell using a
     spell slot of 3rd level or higher, you can target one additional creature for 
    each slot level above 2nd.
    """
    name = "Enhance Ability"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Fur or a feather from a beast"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Cleric', 'Druid', 'Sorcerer')


class Enlargereduce(Spell):
    """You cause a creature or an object you can see within range to grow larger or 
    smaller for the duration. Choose either a creature or an object that is neither 
    worn nor carried. If the target is unwilling, it can make a Constitution saving 
    throw. On a success, the spell has no effect. 
    
    If the target is a creature, 
    everything it is wearing and carrying changes size with it. Any item dropped by 
    an affected creature returns to normal size at once. 
    
    Enlarge 
    The target’s 
    size doubles in all dimensions, and its weight is multiplied by eight. This 
    growth increases its size by one category – from Medium to Large, for example. 
    If there isn’t enough room for the target to double its size, the creature or 
    object attains the maximum possible size in the space available. Until the spell
     ends, the target also has advantage on Strength checks and Strength saving 
    throws. The target’s weapons also grow to match its new size. While these 
    weapons are enlarged, the target’s attack with them deal 1d4 extra damage. 
    
    
    Reduce 
    The target’s size is halved in all dimensions, and its weight is reduced
     to one-eighth of normal. This reduction decreases its size by one category – 
    from Medium to Small, for example. Until the spell ends, the target also has 
    disadvantage on Strength checks and Strength saving throws. The target’s weapons
     also shrink to match its new size. While these weapons are reduced, the 
    target’s attacks with them deal 1d4 less damage (this can’t reduce the damage 
    below 1).
    """
    name = "Enlargereduce"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A pinch of powdered iron"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Wizard')


class EnsnaringStrike(Spell):
    """The next time you hit a creature with a weapon attack before this spell ends, a 
    writhing mass of thorny vines appears at the point of impact, and the target 
    must succeed on a Strength saving throw or be restrained by the magical vines 
    until the spell ends. A Large or larger creature has advantage on this saving 
    throw. If the target succeeds on the save, the vines shrivel away. 
    
    While 
    restrained by this spell, the target takes 1d6 piercing damage at the start of 
    each of its turns. A creature restrained by the vines or one that can touch the 
    creature can use its action to make a Strength check against your spell save DC.
     On a success, the target is freed.
    
    At Higher Levels: If you cast this spell 
    using a spell slot of 2nd level or higher, the damage increases by 1d6 for each 
    slot level above 1st.
    """
    name = "Ensnaring Strike"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Ranger',)


class Entangle(Spell):
    """Grasping weeds and vines sprout from the ground in a 20-foot square starting 
    from a point within range. For the duration, these plants turn the ground in the
     area into difficult terrain. 
    
    A creature in the area when you cast the spell 
    must succeed on a Strength saving throw or be restrained by the entangling 
    plants until the spell ends. A creature restrained by the plants can use its 
    action to make a Strength check against your spell save DC. On a success, it 
    frees itself. 
    
    When the spell ends, the conjured plants wilt away.
    """
    name = "Entangle"
    level = 1
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Druid',)


class Enthrall(Spell):
    """You weave a distracting string of words, causing creatures of your choice that 
    you can see within range and that can hear you to make a Wisdom saving throw. 
    Any creature that can’t be charmed succeeds on this saving throw automatically, 
    and if you or your companions are fighting a creature, it has advantage on the 
    save. On a failed save, the target has disadvantage on Wisdom (Perception) 
    checks made to perceive any creature other than you until the spell ends or 
    until the target can no longer hear you. The spell ends if you are incapacitated
     or can no longer speak.
    """
    name = "Enthrall"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Warlock')


class EruptingEarth(Spell):
    """Choose a point you can see on the ground within range. A fountain of churned 
    earth and stone erupts in a 20-foot cube centered on that point. Each creature 
    in that area must make a Dexterity saving throw. A creature takes 3d12 
    bludgeoning damage on a failed save, or half as much damage on a successful one.
     Additionally, the ground in that area becomes difficult terrain until cleared 
    away. Each 5-foot-square portion of the area requires at least 1 minute to clear
     by hand.
    
    At Higher Levels: When you cast this spell using a spell slot of 4rd 
    level or higher, the damage increases by 1d12 for each slot level above 3rd.
    """
    name = "Erupting Earth"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A piece of obsidian"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class Etherealness(Spell):
    """You step into the border regions of the Ethereal Plane, in the area where it 
    overlaps with your current plane. You remain in the Border Ethereal for the 
    duration or until you use your action to dismiss the spell. During this time, 
    you can move in any direction. If you move up or down, every foot of movement 
    costs an extra foot. You can see and hear the plan you originated from, but 
    everything there looks gray, and you can’t see anything more than 60 feet away.
     
    
    While on the Ethereal Plane, you can only affect and be affected by other 
    creatures on that plane. Creatures that aren’t on the Ethereal Plance can’t 
    perceive you and can’t interact with you, unless a special ability or magic has 
    given them the ability to do so. 
    
    You ignore all objects and effects that 
    aren’t on the Ethereal Plane, allowing you to move through objects you perceive 
    on the plan you originated from. When the spell ends, you immediately return to 
    the plane you originiated from in teh spot you currently occupy. If you occupy 
    the same spot as a solid object or creature when this happens, you are 
    imediately shunted to the neares unoccupied space that you can occupy and take 
    force damage equal to twice the number of feet you are moved. 
    
    This spell has 
    no effect if you cast it while you are on the Ethereal Plane or a plane that 
    doesn’t border it, such as one of the Outer Planes.
    
    At Higher Levels: When you 
    cast this spell using a spell slot of 8th level or higher, you can target up to 
    three willing creatures (including you) for each slot level above 7th. The 
    creatures must be within 10 feet of you when you cast the spell.
    """
    name = "Etherealness"
    level = 7
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "Up to 8 hours"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Cleric', 'Sorcerer', 'Warlock', 'Wizard')


class EvardsBlackTentacles(Spell):
    """Squirming, ebony tentacles fill a 20-foot square on ground that you can see 
    within range. For the duration, these tentacles turn the ground in the area into
     difficult terrain. 
    
    When a creature enters the affected area for the first 
    time on a turn or starts its turn there, the creature must succeed on a 
    Dexterity saving throw or take 3d6 bludgeoning damage and be restrained by the 
    tentacles until the spell ends. A creature that starts its turn in the area and 
    is already restrained by the tentacles takes 3d6 bludgeoning damage. 
    
    A 
    creature restrained by the tentacles can use its action to make a Strength or 
    Dexterity check (its choice) against your spell save DC. On a success, it frees 
    itself.
    """
    name = "Evards Black Tentacles"
    level = 4
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S', 'M')
    materials = """A piece of tentacle from a giant octopus or a giant squid"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Wizard',)


class ExpeditiousRetreat(Spell):
    """This spell allows you to move at an incredible pace. When you cast this spell, 
    and then as a bonus action on each of your turns until the spell ends, you can 
    take the Dash action.
    """
    name = "Expeditious Retreat"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class Eyebite(Spell):
    """For the spell’s duration, your eyes become an inky void imbued with dread power.
     One creature of your choice within 60 feet of you that you can see must succeed
     on a Wisdom saving throw or be affected by one of the following effects of your
     choice for the duration. On each of your turns until the spell ends, you can 
    use your action to target another creature but can’t target a creature again if 
    it has succeeded on a saving throw against this casting of eyebite. 
    
    Asleep 
    
    The target galls unconscious. It wakes up if it takes any damage or if another 
    creature uses its action to shake the sleeper awake. 
    
    Panicked 
    The target is 
    frightened of you. On each of its turns, the frightened creature must take the 
    Dash action and move away from you by the safest and shortest available route, 
    unless there is nowhere to move. If the target moves to a place at least 60 feet
     away from you where it can no longer see you, this effect ends. 
    
    Sickened 
    The
     target has disadvantage on attack rolls and ability checks. At the end of each 
    of its turns, it can make another Wisdom saving throw. If it succeeds, the 
    effect ends.
    """
    name = "Eyebite"
    level = 6
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')





from .spells import Spell


class SacredFlame(Spell):
    """Flame-like radiance descends on a creature that you can see within range. The 
    target must succeed on a Dexterity saving throw or take 1d8 radiant damage. The 
    target gains no benefit from cover for this saving throw.
    
    At Higher Levels: The
     spell’s damage increases by 1d8 when you reach 5th level (2d8), 11th level 
    (3d8), and 17th level (4d8).
    """
    name = "Sacred Flame"
    level = 0
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric',)


class Sanctuary(Spell):
    """You ward a creature within range against attack.
    Until the spell ends, any 
    creature who targets the warded creature with an attack or a harmful spell must 
    first make a Wisdom saving throw. On a failed save, the creature must choose a 
    new target or lose the attack or spell. This spell doesn’t protect the warded 
    creature from area effects, such as the explosion of a fireball.
    
    If the warded 
    creature makes an attack or casts a spell that affects an enemy creature, this 
    spell ends.
    """
    name = "Sanctuary"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "30 feet"
    components = ('V', 'S', 'M')
    materials = """A small silver mirror"""
    duration = "1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric',)


class Scatter(Spell):
    """The air quivers around up to five creatures of your choice that you can see 
    within range. An unwilling creature must succeed on a Wisdom saving throw to 
    resist this spell. You teleport each affected target to an unoccupied space that
     you can see within 120 feet of you. That space must be on the ground or on a 
    floor.
    """
    name = "Scatter"
    level = 6
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class ScorchingRay(Spell):
    """You create three rays of fire and hurl them at targets within range. You can 
    hurl them at one target or several. Make a ranged spell attack for each ray. On 
    a hit, the target takes 2d6 fire damage.
    
    At Higher Levels: When you cast this 
    spell using a spell slot of 3rd level or higher, you create one additional ray 
    for each slot level above 2nd.
    """
    name = "Scorching Ray"
    level = 2
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class Scrying(Spell):
    """You can see and hear a particular creature you choose that is on the same plane 
    of existence as you. The target must make a W isdom saving throw, which is 
    modified by how well you know the target and the sort of physical connection you
     have to it. If a target knows you’re casting this spell, it can fail the saving
     throw voluntarily if it wants to be observed.
    
    Knowledge                 Save 
    Modifier
    Secondhand (you have heard of the target) +5
    Firsthand (you have met 
    the target)      +0
    Familiar (you know the target well)     -5
    
    Connection 
                   Save Modifier
    Likeness or picture               -2
    Posession or 
    garment            -4
    Body part, lock of hair, bit of nail, or the like -10
    
    On 
    a successful save, the target isn’t affected, and you can’t use this spell 
    against it again for 24 hours.
    
    On a failed save, the spell creates an invisible
     sensor within 10 feet of the target. You can see and hear through the sensor as
     if you w ere there. The sensor moves with the target, remaining within 10 feet 
    of it for the duration. A creature that can see invisible objects sees the 
    sensor as a luminous orb about the size of your fist.
    
    Instead of targeting a 
    creature, you can choose a location you have seen before as the target of this 
    spell. When you do, the sensor appears at that location and doesn’t move.
    """
    name = "Scrying"
    level = 5
    casting_time = "10 minutes"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """A focus worth at least 1,000 gp, such as a crystal ball, a silver mirror, or a font filled with holy water"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Divination"
    classes = ('Bard', 'Cleric', 'Druid', 'Warlock', 'Wizard')


class SearingSmite(Spell):
    """The next time you hit a creature with a melee weapon attack during the spell’s 
    duration, your weapon flares with white-hot intensitity, and the attack deals an
     extra 1d6 fire damage to the target and causes the target to ignite in flames.
    
    
    At the start of each of its turns until the spell ends, the target must make a
     Constitution saving throw. On a failed save, it takes 1d6 fire damage. On a 
    successful save, the spells ends. If the target or a creature within 5 feet of 
    it uses an action to put out the flames, or if some other effect douses the 
    flames (such as the target being submerged in water), the spell ends.
    
    At Higher
     Levels: When you cast this spell using a spell slot of 2nd level or higher, the
     initial extra damage dealt by the attack increases by 1d6 for each slot
    """
    name = "Searing Smite"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Paladin',)


class SeeInvisibility(Spell):
    """For the duration, you see invisible creatures and objects as if they were 
    visible, and you can see into the Ethereal Plane. Ethereal creatures and objects
     appear ghostly and translucent.
    """
    name = "See Invisibility"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """A pinch of talc and a small sprinkling of powdered silver"""
    duration = "1 hour"
    ritual = False
    magic_school = "Divination"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class Seeming(Spell):
    """This spell allows you to change the appearance of any number of creatures that 
    you can see within range.
    You give each target you choose a new, illusory 
    appearance. An unwilling target can make a Charisma saving throw, and if it 
    succeeds, it is unaffected by this spell.
    
    The spell disguises physicial 
    appearances as well as clothing, armor, weapons, and equipment. You can make 
    each creature seem 1 foot shorter or taller and appear thin, fat, or inbetween. 
    You can’t change a target’s body type, so you must choose a form that has the 
    same basic arrangement of limbs. Otherwise, the extent of the illusion is up to 
    you. The spell lasts for the duration, unless you use your action to dismiss it 
    sooner.
    
    The changes wrought by this spell fail to hold up to physical 
    inspections. For example, if you use this spell to add a hat to a creature’s 
    outfitm objects pass through the hat, and anyone who touches it would feel 
    nothing or would feel the creature’s head and hair. If you use this spell to 
    appear thinner then you are, the hand of someone who reaches out to touch you 
    would bump into you while it was seemingly still in midair.
    
    A creature can use 
    its action to inspect a target and make an Intelligence (Investigation) check 
    against your spell save DC. If it succeeds, it becomes aware that the target is 
    disguised.
    """
    name = "Seeming"
    level = 5
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "8 hours"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class Sending(Spell):
    """You send a short message of twenty-five words or less to a creature with you are
     familiar. The creature hears the message in its mind, regonizes you as the 
    sender if it knows you, and can answer in a like manner immediately. The spell 
    enables creatures with Intelligence scores of at least 1 to understand the 
    meaning of your message.
    
    You can send the message across any distance and even 
    to other planes of existence, but if the target is on a different plane than 
    you, there is a 5 percent chance that the message doesn’t arrive.
    """
    name = "Sending"
    level = 3
    casting_time = "1 action"
    casting_range = "Unlimited"
    components = ('V', 'S', 'M')
    materials = """A short piece of fine copper wire"""
    duration = "1 round"
    ritual = False
    magic_school = "Evocation"
    classes = ('Bard', 'Cleric', 'Wizard')


class Sequester(Spell):
    """By means of this spell, a willing creature of an object can be hidden away, safe
     from detection for the duration.
    When you cast the spell and touch the target, 
    it becomes invisible and can’t be targeted by divination spells or perceived 
    through scrying sensors created by the divination of spells.
    
    If the target is a
     crreature, it falls into a state of suspended animation. Time ceases to flow 
    for it, and it doesn’t grow older.
    
    You can set a condition for the spell to end
     early. The condition can be anything you choose, but it must occur or be 
    visible within 1 mile of the target. Examples include after 1,000 years or when 
    the tarrasque awakes. This spells also ends if the target takes any damage.
    """
    name = "Sequester"
    level = 7
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A powder composed of diamond, emerald, ruby, and sapphire dust worth at least 5,000 gp, which the spell consumes"""
    duration = "Until dispelled"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Wizard',)


class ShadowBlade(Spell):
    """You weave together threads of shadow to create a sword of solidified gloom in 
    your hand. This magic sword lasts until the spell ends. It counts as a simple 
    melee weapon with which you are proficient. It deals 2d8 psychic damage on a hit
     and has the finesse, light, and thrown properties (range 20/60). In addition, 
    when you use the sword to attack a target that is in dim light or darkness, you 
    make the attack roll with advantage.
    If you drop the weapon or throw it, it 
    dissipates at the end of the turn. Thereafter, while the spell persists, you can
     use a bonus action to cause the sword to reappear in your hand.
    
    At Higher 
    Levels: When you cast this spell using a 3rd- or 4th-level spell slot, the 
    damage increases to 3d8. When you cast it using a 5th- or 6th-level spell slot, 
    the damage increases to 4d8. When you cast it using a spell slot of 7th level or
     higher, the damage increases to 5d8.
    """
    name = "Shadow Blade"
    level = 2
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class ShadowOfMoil(Spell):
    """Flame-like shadows wreathe your body until the spell ends, causing you to become
     heavily obscured to others. The shadows turn dim light within 10 feet of you 
    into darkness, and bright light in the same area to dim light.
    Until the spell 
    ends, you have resistance to radiant damage. In addition, whenever a creature 
    within 10 feet of you hits you with an attack, the shadows lash out at that 
    creature, dealing it 2d8 necrotic damage.
    """
    name = "Shadow Of Moil"
    level = 4
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """An undead eyeball encased in a gem worth at least 150 gp"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Warlock',)


class ShapeWater(Spell):
    """You choose an area of water that you can see within range and that fits within a
     5-foot cube. 
    You manipulate it in one of the following ways:
    
    • You 
    instantaneously move or otherwise change the flow of the water as you direct, up
     to 5 feet in any direction. This movement doesn’t have enough force to cause 
    damage.
    
    • You cause the water to form into simple shapes and animate at your 
    direction. This change lasts for 1 hour.
    
    • You change the water’s color or 
    opacity. The water must be changed in the same way throughout. This change lasts
     for 1 hour.
    
    • You freeze the water, provided that there are no crea- tures in 
    it. The water unfreezes in 1 hour.
    If you cast this spell multiple times, you 
    can have no more than two of its non-instantaneous effects active at a time, and
     you can dismiss such an effect as an action.
    """
    name = "Shape Water"
    level = 0
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('S',)
    materials = """"""
    duration = "Instantaneous or 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class Shapechange(Spell):
    """You assume the form of a different creature for the duration.
    The new form can 
    be any creature with a challenge rating equal to your level or lower. The 
    creature can’t be a construct or an undead, and you must have seen the sort of 
    creature at least once. You transform into an average example of that creature, 
    one without any class levels or the Spellcasting trait.
    
    Your game statistics 
    are replaced by the statistics of the chosen creature, though you retain your 
    alignment and Intelligence, Wisdom, and Charisma scores You also retain all of 
    your skill and saving throw proficiencies, in addition to gaining those of the 
    creature, If the creature has the same proficiency as you, and the bonus listed 
    in its statistics is higher than yours, use the creature’s bonus in place of 
    yours. You can’t use any legendary actions or lair actions of the new form.
    
    You
     assume the hit points and Hit Dice of the new form. When you revert to your 
    normal, you return to the number of hit points you had before you transformed. 
    If you revert as a result of dropping to 0 hit points, any excess damage carries
     over to your normal form. As long as the excess damage doesn’t reduce your 
    normal form to 0 hitpoints, you aren’t knocked unconscious.
    
    You retain the 
    benefit of any features from your class, race, or other source and can use them,
     provided that your new form is physically capable of doing so. You can’t use 
    any special senses you have (for example, darkvision) unless your new form also 
    has that sense. You can only speak if the creature can normally speak.
    
    When you
     transform, you choose whether your equipment falls to the ground, merges into 
    the new form, or is worn by it. Worn equipment functions as normal. The DM 
    determines whether it is practical for the new form to wear a piece of 
    equipment, based on the creature’s shape and size. Your equipment doesn’t change
     shape or size to match the new form, and any equipment that the new form can’t 
    wear must either fall to the ground or merge into your new form. Equipment that 
    merges has no effect in that state.
    
    During this spell’s duration, you can use 
    your action to assume a different form following the same restrictions and rules
     for the original form, with one exception: if your new form has more hit pints 
    than your current one, your hit points remain at their current value.
    """
    name = "Shapechange"
    level = 9
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S', 'M')
    materials = """A jade circlet worth at least 1,500 gp, which you must place on your head before you cast the spell"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Wizard')


class Shatter(Spell):
    """A sudden loud ringing noise, painfully intense, erupts from a point of your 
    choice within range. Each creature in a 10-foot-radius sphere centered on that 
    point must make a Constitution saving throw. A creature takes 3d8 thunder damage
     on a failed save, or half as much damage on a successful one. A creature made 
    of inorganic material such as stone, crystal, or metal has disadvantage on this 
    saving throw.
    
    A nonmagical object that isn’t being worn or carried also takes 
    the damage if it’s in the spell’s area.
    
    At Higher Levels: When you cast this 
    spell using a spell slot of 3rd level or higher, the damage increases by 1d8 for
     each slot level above 2nd.
    """
    name = "Shatter"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A chip of mica"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class Shield(Spell):
    """Reaction trigger: You are hit by an attack or targeted by the magic missile 
    spell
    
    An invisible barrier of magical force appears and protects you. Until the
     start of your next turn, you have a +5 bonus to AC, including against the 
    triggering attack, and you take no damage from magic missile.
    """
    name = "Shield"
    level = 1
    casting_time = "Special"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "1 round"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Sorcerer', 'Wizard')


class ShieldOfFaith(Spell):
    """A shimmering field appears and surrounds a creature of your choice within range,
     granting it a +2 bonus to AC for the duration.
    """
    name = "Shield Of Faith"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A small parchment with a bit of holy text written on it"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Cleric', 'Paladin')


class Shillelagh(Spell):
    """The wood of a club or quarterstaff you are holding is imbued with nature’s 
    power.
    For the duration, you can use your spellcasting ability instead of 
    Strength for the attack and damage rolls of melee attacks using that weapon, and
     the weapon’s damage die becomes a d8. The weapon also becomes magical, if it 
    isn’t already. The spell ends if you cast it again or if you let go of the 
    weapon
    """
    name = "Shillelagh"
    level = 0
    casting_time = "1 bonus action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Mistletoe, a shamrock leaf, and a club or quarterstaff"""
    duration = "1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid',)


class ShockingGrasp(Spell):
    """Lightning springs from your hand to deliver a shock to a creature you try to 
    touch.
    Make a melee spell attack against the target. You have advantage on the 
    attack roll if the target is wearing armor made of metal. On a hit, the target 
    takes 1d8 lightning damage, and it can’t take reactions until the start of its 
    next turn.
    
    At Higher Levels: The spell’s damage increases by 1d8 when you reach
     5th level (2d8), 11th level (3d8), and 17th level (4d8).
    """
    name = "Shocking Grasp"
    level = 0
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class SickeningRadiance(Spell):
    """Dim, greenish light spreads within a 30-foot-radius sphere centered on a point 
    you choose within range. The light spreads around corners, and it lasts until 
    the spell ends.
    When a creature moves into the spell’s area for the first time 
    on a turn or starts its turn there, that creature must succeed on a Constitution
     saving throw or take 4d10 radiant damage, and it suffers one level of 
    exhaustion and emits a dim, greenish light in a 5-foot radius. This light makes 
    it impossible for the creature to benefit from being invisible. The light and 
    any levels of exhaustion caused by this spell go away when the spell ends.
    """
    name = "Sickening Radiance"
    level = 4
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class Silence(Spell):
    """For the duration, no sound can be created within or pass through a 20-foot-
    radius sphere centered on a point you choose within range. Any creature or 
    object entirely inside the sphere is immune to thunder damage, and creatures are
     deafened while entirely inside it. Casting a spell that includes a verbal 
    component is impossible there.
    """
    name = "Silence"
    level = 2
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 10 minutes"
    ritual = True
    magic_school = "Illusion"
    classes = ('Bard', 'Cleric', 'Ranger')


class SilentImage(Spell):
    """You create the image of an object, a creature, or some other visible phenomenon 
    that is no larger than a 15-foot cube. The image appears at a spot within range 
    and lasts for the duration. The image is purely visual; it isn’t accompanied by 
    sound, smell, or other sensory effects.
    
    You can use your action to cause the 
    image to move to any spot within range. As the image changes location, you can 
    alter its appearance so that its movements appear natural for the image. For 
    example, if you create an image of a creature and move it, you can alter the 
    image so that it appears to be walking.
    
    Physical interaction with the image 
    reveals it to be an illusion, because things can pass through it. A creature 
    that uses its action to examine the image can determine that it is an illusion 
    with a successful Intelligence (Investigation) check against your spell save DC.
     If a creature discerns the illusion for what it is, the creature can see 
    through the image.
    """
    name = "Silent Image"
    level = 1
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A bit of fleece"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Illusion"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class Simulacrum(Spell):
    """You shape an illusory duplicate of one beast or humanoid that is within range 
    for the entire casting time of the spell.
    The duplicate is a creature, partially
     real and formed from ice or snow, and it can take actions and otherwise be 
    affected as a normal creature. It appears to be the same as the original, but it
     has half the creature’s hit point maximum and is formed without any equipment. 
    Otherwise, the illusion uses all the statistics of the creature it duplicates.
    
    
    The simulacrum is friendly to you and creatures you designate. It obeys your 
    spoken commands, moving and acting in accordance with your wishes and acting on 
    your turn in combat. The simulacrum lacks the ability to learn or become more 
    powerful, so it never increases its level or other abilities, nor can it regain 
    expended spell slots.
    
    If the simulacrum is damaged, you can repair it in an 
    alchemical laboratory, using rare herbs and minerals worth 100 gp per hit point 
    it regains. The simulacrum lasts until it drops to 0 hit points, at which point 
    it reverts to snow and melts instantly.
    
    If you cast this spell again, any 
    currently active duplicates you created with this spell are instantly destroyed.
    """
    name = "Simulacrum"
    level = 7
    casting_time = "12 hours"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Snow or ice in quantities sufficient to made a life-size copy of the duplicated creature; some hair, fingernail clippings, or other piece of that creature’s body placed inside the snow or ice; and powdered ruby worth 1,500 gp, sprinkled over the duplicate and consumed by the spell"""
    duration = "Until dispelled"
    ritual = False
    magic_school = "Illusion"
    classes = ('Wizard',)


class SkillEmpowerment(Spell):
    """Your magic deepens a creature’s understanding of its own talent. You touch one 
    willing creature and give it expertise in one skill of your choice; until the 
    spell ends, the creature doubles its proficiency bonus for ability checks it 
    makes that use the chosen skill.
    You must choose a skill in which the target is 
    proficient and that isn’t already benefiting from an effect, such as Expertise, 
    that doubles its proficiency bonus.
    """
    name = "Skill Empowerment"
    level = 5
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class Skywrite(Spell):
    """You cause up to ten words to form in a part of the sky you can see. The words 
    appear to be made of cloud and remain in place for the spell’s duration. The 
    words dissipate when the spell ends. A strong wind can disperse the clouds and 
    end the spell early.
    """
    name = "Skywrite"
    level = 2
    casting_time = "1 action"
    casting_range = "Sight"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 hour"
    ritual = True
    magic_school = "Transmutation"
    classes = ('Bard', 'Druid', 'Wizard')


class Sleep(Spell):
    """This spell sends creatures into a magical slumber. Roll 5d8, the total is how 
    many hit points of creatures this spell can affect. Creatures within 20 feet of 
    a point you choose within range are affected in ascending order of their current
     hit points (ignoring unconscious creatures).
    
    Starting with the creature that 
    has the lowest current hit points, each creature affected by this spell falls 
    unconscious until the spell ends, the sleeper takes damage, or someone uses an 
    action to shake or slap the sleeper awake. Subtract each creature’s hit points 
    from the total before moving on to the creature with the next lowest hit points.
     A creature’s hit points must be equal to or less than the remaining total for 
    that creature to be affected.
    Undead and creatures immune to being charmed 
    aren’t affected by this spell.
    
    At Higher Levels: When you cast this spell using
     a spell slot of 2nd level or higher, roll an additional 2d8 for each slot level
     above 1st.
    """
    name = "Sleep"
    level = 1
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S', 'M')
    materials = """A pinch of fine sand, rose petals, or a cricket"""
    duration = "1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class SleetStorm(Spell):
    """Until the spell ends, freezing rain and sleet fall in a 20-foot-tall cylinder 
    with a 40-foot radius centered on a point you choose within range. The area is 
    heavily obscured, and exposed flames in the area are doused.
    
    The ground in the 
    area is covered with slick ice, making it difficult terrain. When a creature 
    enters the spell’s area for the first time on a turn or starts its turn there, 
    it must make a Dexterity saving throw. On a failed save, it falls prone.
    
    If a 
    creature is concentrating in the spell’s area, the creature must make a 
    successful Constitution saving throw against your spell save DC or lose 
    concentration.
    """
    name = "Sleet Storm"
    level = 3
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ('V', 'S', 'M')
    materials = """A pinch of dust and a few drops of water"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class Slow(Spell):
    """You alter time around up to six creatures of your choice in a 40-foot cube 
    within range. Each target must succeed on a Wisdom saving throw or be affected 
    by this spell for the duration.
    
    An affected target’s speed is halved, it takes 
    a -2 penalty to AC and Dexterity saving throws, and it can’t use reactions. On 
    its turn, it can use either an action or a bonus action, not both. Regardless of
     the creature’s abilities or magic items, it can’t make more than one melee or 
    ranged attack during its turn.
    
    If the creature attempts to cast a spell with a 
    casting time of 1 action, roll a d20. On an 11 or higher, the spell doesn’t take
     effect until the creature’s next turn, and the creature must use its action on 
    that turn to complete the spell. If it can’t, the spell is wasted.
    
    A creature 
    affected by this spell makes another Wisdom saving throw at the end of its turn.
     On a successful save, the effect ends for it.
    """
    name = "Slow"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S', 'M')
    materials = """A drop of molasses"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Wizard')


class Snare(Spell):
    """As you cast this spell, you use the rope to create a circle with a 5-foot radius
     on the ground or the floor. When you finish casting, the rope disappears and 
    the circle becomes a magic trap.
    This trap is nearly invisible, requiring a 
    successful Intelligence (Investigation) check against your spell save DC to be 
    discerned.
    The trap triggers when a Small, Medium, or Large creature moves onto 
    the ground or the floor in the spell’s radius. That creature must succeed on a 
    Dexterity saving throw or be magically hoisted into the air, leaving it hanging 
    upside down 3 feet above the ground or the floor. The creature is restrained 
    there until the spell ends.
    A restrained creature can make a Dexterity saving 
    throw at the end of each of its turns, ending the effect on itself on a success.
     Alternatively, the creature or someone else who can reach it can use an action 
    to make an Intelligence (Arcana) check against your spell save DC. On a success,
     the restrained effect ends.
    After the trap is triggered, the spell ends when no
     creature is restrained by it.
    """
    name = "Snare"
    level = 1
    casting_time = "1 minute"
    casting_range = "Touch"
    components = ('S', 'M')
    materials = """25 feet of rope, which the spell consumes"""
    duration = "8 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Druid', 'Ranger', 'Wizard')


class SnillocsSnowballSwarm(Spell):
    """(a piece of ice or a small white rock chip)
    A flurry of magic snowballs erupts 
    from a point you choose within range. Each creature in a 5-foot-radius sphere 
    centered on that point must make a Dexterity saving throw. A creature takes 3d6 
    cold damage on a failed save, or half as much damage on a successful one.
    At 
    Higher Levels. When you cast this spell using a spell slot of 3rd level or 
    higher, the damage increases by 1d6 for each slot level above 2nd.
    """
    name = "Snillocs Snowball Swarm"
    level = 2
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S', 'M')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class SoulCage(Spell):
    """1 reaction, which you take when a humanoid you can see within 60 feet of you 
    dies
    
    This spell snatches the soul of a humanoid as it dies and traps it inside 
    the tiny cage you use for the material component. A stolen soul remains inside 
    the cage until the spell ends or until you destroy the cage, which ends the 
    spell. While you have a soul inside the cage, you can exploit it in any of the 
    ways described below. You can use a trapped soul up to six times. Once you 
    exploit a soul for the sixth time, it is released, and the spell ends. While a 
    soul is trapped, the dead humanoid it came from can’t be revived.
    Steal Life. 
    You can use a bonus action to drain vigor from the soul and regain 2d8 hit 
    points.
    Query Soul. You ask the soul a question (no action required) and receive
     a brief telepathic answer, which you can understand regardless of the language 
    used. The soul knows only what it knew in life, but it must answer you 
    truthfully and to the best of its ability. The answer is no more than a sentence
     or two and might be cryptic.
    Borrow Experience. You can use a bonus action to 
    bolster yourself with the soul’s life experience, making your next attack roll, 
    ability check, or saving throw with advantage. If you don’t use this benefit 
    before the start of your next turn, it is lost.
    Eyes of the Dead. You can use an
     action to name a place the humanoid saw in life, which creates an invisible 
    sensor somewhere in that place if it is on the plane of existence you’re 
    currently on. The sensor remains for as long as you concentrate, up to 10 
    minutes (as if you were concentrating on a spell). You receive visual and 
    auditory information from the sensor as if you were in its space using your 
    senses.
    A creature that can see the sensor (such as one using see invisibility 
    or truesight) sees a translucent image of the tormented humanoid whose soul you 
    caged.
    """
    name = "Soul Cage"
    level = 6
    casting_time = "Special"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A tiny silver cage worth 100 gp"""
    duration = "8 hours"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Warlock', 'Wizard')


class SpareTheDying(Spell):
    """You touch a living creature that has 0 hit points. The creature becomes stable. 
    This spell has no effect on undead or constructs.
    """
    name = "Spare The Dying"
    level = 0
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Cleric',)


class SpeakWithAnimals(Spell):
    """You gain the ability to comprehend and verbally communicate with beasts for the 
    duration.
    The knowledge and awareness of many beasts is limited by their 
    intelligence, but at minimum, beasts can give you information about nearby 
    locations and monsters, including whatever they can perceive or have perceived 
    within the past day. You might be able to persuade a beast to perform a small 
    favor for you, at the DM’s discretion.
    """
    name = "Speak With Animals"
    level = 1
    casting_time = "1 action"
    casting_range = "Self"
    components = ('V', 'S')
    materials = """"""
    duration = "10 minutes"
    ritual = True
    magic_school = "Divination"
    classes = ('Bard', 'Druid', 'Ranger')


class SpeakWithDead(Spell):
    """You grant the semblance of life and intelligence to a corpse of your choice 
    within range, allowing it to answer the questions you pose. The corpse must 
    still have a mouth and can’t be undead. The spell fails if the corpse was the 
    target of this spell within the last 10 days.
    
    Until the spell ends, you can ask
     the corpse up to five questions. The corpse knows only what it knew in life, 
    including the languages it knew. Answers are usually brief, cryptic, or 
    repetitive, and the corpse is under no compulsion to offer a truthful answer if 
    you are hostile to it or it recognizes you as an enemy. This spell doesn’t 
    return the creature’s soul to its body, only its animating spirit. Thus, the 
    corpse can’t learn new information, doesn’t comprehend anything that has 
    happened since it died, and can’t speculate about future events.
    """
    name = "Speak With Dead"
    level = 3
    casting_time = "1 action"
    casting_range = "10 feet"
    components = ('V', 'S', 'M')
    materials = """Burning incense"""
    duration = "10 minutes"
    ritual = False
    magic_school = "Necromancy"
    classes = ('Bard', 'Cleric')


class SpeakWithPlants(Spell):
    """You imbue plants within 30 feet of you with limited sentience and animation, 
    giving them the ability to communicate with you and follow your simple commands.
     You can question plants about events in the spell’s area within the past day, 
    gaining information about creatures that have passed, weather, and other 
    circumstances.
    
    You can also turn difficult terrain caused by plant growth (such
     as thickets and undergrowth) into ordinary terrain that lasts for the duration.
     Or you can turn ordinary terrain where plants are present into difficult 
    terrain that lasts for the duration, causing vines and branches to hinder 
    pursuers, for example.
    
    Plants might be able to perform other tasks on your 
    behalf, at the DM’s discretion. The spell doesn’t enable plants to uproot 
    themselves and move about, but they can freely move branches, tendrils, and 
    stalks.
    
    If a plant creature is in the area, you can communicate with it as if 
    you shared a common language, but you gain no magical ability to influence it.
    
    
    This spell can cause the plants created by the entangle spell to release a 
    restrained creature.
    """
    name = "Speak With Plants"
    level = 3
    casting_time = "1 action"
    casting_range = "Self (30-foot radius)"
    components = ('V', 'S')
    materials = """"""
    duration = "10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Bard', 'Druid', 'Ranger')


class SpiderClimb(Spell):
    """Until the spell ends, one willing creature you touch gains the ability to move 
    up, down, and across vertical surfaces and upside down along ceilings, while 
    leaving its hands free. The target also gains a climbing speed equal to its 
    walking speed.
    """
    name = "Spider Climb"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A drop of bitumen and a spider"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class SpikeGrowth(Spell):
    """The ground in a 20-foot radius centered on a point within range twists and 
    sprouts hard spikes and thorns. The area becomes difficult terrain for the 
    duration. When a creature moves into or within the area, it takes 2d4 piercing 
    damage for every 5 feet it travels.
    
    The transformation of the ground is 
    camouflaged to look natural. Any creature that can’t see the area at the time 
    the spell is case must make a Wisdom (Perception) check against your spell save 
    DC to recognize the terrain as hazardous before entering it.
    """
    name = "Spike Growth"
    level = 2
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ('V', 'S', 'M')
    materials = """Seven sharp thorns or seven small twigs, each sharpened to a point"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Druid', 'Ranger')


class SpiritGuardians(Spell):
    """You call forth spirits to protect you.
    They flit around you to a distance of 15 
    feet for the duration. If you are good or neutral, their spectral form appears 
    angelic or fey (your choice). If you are evil, they appear fiendish.
    
    When you 
    cast this spell, you can designate any number of creatures you can see to be 
    unaffected by it. An affected creature’s speed is halved in the area, and when 
    the creature enters the area for the first time on a turn or starts its turn 
    there, it must make a Wisdom saving throw. On a failed save, the creature takes 
    3d8 radiant damage (if you are good or neutral) or 3d8 necrotic damage (if you 
    are evil). On a successful save, the creature takes half as much damage.
    
    At 
    Higher Levels: When you cast this spell using a spell slot of 4th level or 
    higher, the damage increases by 1d8 for each slot level above 3rd.
    """
    name = "Spirit Guardians"
    level = 3
    casting_time = "1 action"
    casting_range = "Self (15-foot-radius)"
    components = ('V', 'S', 'M')
    materials = """A holy symbol"""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Cleric',)


class SpiritualWeapon(Spell):
    """You create a floating, spectral weapon within range that lasts for the duration 
    or until you cast this spell again.
    When you cast the spell, you can make a 
    melee spell attack against a creature within 5 feet of the weapon. On a hit, the
     target takes force damage equal to 1d8 + your spellcasting ability modifier.
    
    
    As a bonus action on your turn, you can move the weapon up to 20 feet and repeat
     the attack against a creature within 5 feet of it.
    
    The weapon can take 
    whatever form you choose. Clerics of deities who are associated with a 
    particular weapon (as St. Cuthbert is known for his mace and Thor for his 
    hammer) make this spell’s effect resemble that weapon.
    
    At Higher Levels: When 
    you cast this spell using a spell slot of 3rd level or higher, the damage 
    increases by 1d8 for every two slot levels above the 2nd.
    """
    name = "Spiritual Weapon"
    level = 2
    casting_time = "1 bonus action"
    casting_range = "60 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Cleric',)


class StaggeringSmite(Spell):
    """The next time you hit a creature with a melee weapon attack during this spell’s 
    duration, your weapon pierces both body and mind, and the attack deals an extra 
    4d6 psychic damage to the target. The target must make a Wisdom saving throw. On
     a failed save, it has disadvantage on attack rolls and ability checks, and 
    can’t take reactions, until the end of its next turn.
    """
    name = "Staggering Smite"
    level = 4
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ('V',)
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Paladin',)


class SteelWindStrike(Spell):
    """You flourish the weapon used in the casting and then vanish to strike like the 
    wind. Choose up to five creatures you can see within range. Make a melee spell 
    attack against each target. On a hit, a target takes 6d10 force damage.
    You can 
    then teleport to an unoccupied space you can see within 5 feet of one of the 
    targets you hit or missed.
    """
    name = "Steel Wind Strike"
    level = 5
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('S', 'M')
    materials = """A melee weapon worth at least 1 sp"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Ranger', 'Wizard')


class StinkingCloud(Spell):
    """You create a 20-foot-radius sphere of yellow, nauseating gas centered on a point
     within range. The cloud spreads around corners, and its area is heavily 
    obscured. The cloud lingers in the air for the duration.
    
    Each creature that is 
    completely within the cloud at the start of its turn must make a Constitution 
    saving throw against poison. On a failed save, the creature spends its action 
    that turn retching and reeling. Creatures that don’t need to breathe or are 
    immune to poison automatically succeed on this saving throw.
    
    A moderate wind 
    (at least 10 miles per hour) disperses the cloud after 4 rounds. A strong wind 
    (at least 20 miles per hour) disperses it after 1 round.
    """
    name = "Stinking Cloud"
    level = 3
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ('V', 'S', 'M')
    materials = """A rotten egg or several skunk cabbage leaves"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Bard', 'Sorcerer', 'Wizard')


class StoneShape(Spell):
    """You touch a stone object of Medium size or smaller or a section of stone no more
     than 5 feet in any dimension and form it into any shape that suits your 
    purpose. So, for example, you could shape a large rock into a weapon, idol, or 
    coffer, or make a small passage through a wall, as long as the wall is less than
     5 feet thick. You could also shape a stone door or its frame to seal the door 
    shut. The object you create can have up to two hinges and a latch, but finer 
    mechanical detail isn’t possible.
    """
    name = "Stone Shape"
    level = 4
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Soft clay, which must be worked into roughly the desired shape of the stone object"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Cleric', 'Druid', 'Wizard')


class Stoneskin(Spell):
    """This spell turns the flesh of a willing creature you touch as hard as stone. 
    Until the spell ends, the target has resistance to nonmagical bludgeoning, 
    piercing, and slashing damage.
    """
    name = "Stoneskin"
    level = 4
    casting_time = "1 action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Diamond dust worth 100 gp, which the spell consumes"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Druid', 'Ranger', 'Sorcerer', 'Wizard')


class StormOfVengeance(Spell):
    """A churning storm cloud forms, centered on a point you can see and spreading to a
     radius of 360 feet. Lightning flashes in the area, thunder booms, and strong 
    winds roar. Each creature under the cloud (no more than 5,000 feet beneath the 
    cloud) when it appears must make a Constitution saving throw. On a failed save, 
    a creature takes 2d6 thunder damage and becomes deafened for 5 minutes.
    
    Each 
    round you maintain concentration on this spell, the storm produces additional 
    effects on your turn.
    
    Round 2
    Acidic rain falls from the cloud. Each creature 
    and object under the cloud takes 1d6 acid damage.
    
    Round 3You call six bolts of 
    lightning from the cloud to strike six creatures or objects of your choice 
    beneath the cloud. A given creature or object can’t be struck by more than one 
    bolt. A struck creature must make a Dexterity saving throw. The creature takes 
    10d6 lightning damage on a failed save, or half as much damage on a successful 
    one.
    
    Round 4
    Hailstones rain down from the cloud. Each creature under the cloud
     takes 2d6 bludgeoning damage.
    
    Round 5-10
    Gusts and freezing rain assail the 
    area under the cloud. the area becomes difficult terrain and is heavily 
    obscured. Each creature there takes 1d6 cold damage. Ranged weapon attacks in 
    the area are impossible. The wind and rain count as a severe distraction for the
     purposes of maintaining concentration on spells. Finally, gusts of strong wind 
    (ranging from 20 to 50 miles per hour) automatically disperse fog, mists, and 
    similar phenomena in the area whether mundane or magical.
    """
    name = "Storm Of Vengeance"
    level = 9
    casting_time = "1 action"
    casting_range = "Sight"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Druid',)


class StormSphere(Spell):
    """A 20-foot-radius sphere of whirling air springs into existence centered on a 
    point you choose within range. The sphere remains for the spell’s duration. Each
     creature in the sphere when it appears or that ends its turn there must succeed
     on a Strength saving throw or take 2d6 bludgeoning damage. The sphere’s space 
    is difficult terrain.
    Until the spell ends, you can use a bonus action on each 
    of your turns to cause a bolt of lightning to leap from the center of the sphere
     toward one creature you choose within 60 feet of the center. Make a ranged 
    spell attack. You have advantage on the attack roll if the target is in the 
    sphere. On a hit, the target takes 4d6 lightning damage.
    Creatures within 30 
    feet of the sphere have disadvantage on Wisdom (Perception) checks made to 
    listen.
    At Higher Levels. When you cast this spell using a spell slot of 5th 
    level or higher, the damage increases for each of its effects by 1d6 for each 
    slot level above 4th.
    """
    name = "Storm Sphere"
    level = 4
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Sorcerer', 'Wizard')


class Suggestion(Spell):
    """You suggest a course of activity (limited to a sentence or two) and magically 
    influence a creature you can see within range that can hear and understand you. 
    Creatures that can’t be charmed are immune to this effect. The suggestion must 
    be worded in such a manner as to make the course of action sound reasonable. 
    Asking the creature to stab itself, throw itself onto a spear, immolate itself, 
    or do some other obviously harmful act ends the spell.
    
    The target must make a 
    Wisdom saving throw. On a failed save, it purses the course of action you 
    described to the best of its ability. The suggested course of action can 
    continue for the entire duration. If the suggested activity can be completed in 
    a shorter time, the spell ends when the subject finishes what it was asked to 
    do.
    
    You can also specify conditions that will trigger a special activity during
     the duration. For example, you might suggest that a knight give her warhorse to
     the first beggar she meets. If the condition isn’t met before the spell 
    expires, the activity isn’t preformed.
    
    If you or any of your companions damage 
    the target, the spell ends.
    """
    name = "Suggestion"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ('V', 'M')
    materials = """A snake’s tongue and either a bit of honeycomb or a drop of sweet oil"""
    duration = "Concentration, up to 8 hours"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')


class SummonGreaterDemon(Spell):
    """You utter foul words, summoning one demon from the chaos of the Abyss. You 
    choose the demon’s type, which must be one of challenge rating 5 or lower, such 
    as a shadow demon or a barlgura. The demon appears in an unoccupied space you 
    can see within range, and the demon disappears when it drops to 0 hit points or 
    when the spell ends.
    Roll initiative for the demon, which has its own turns. 
    When you summon it and on each of your turns thereafter, you can issue a verbal 
    command to it (requiring no action on your part), telling it what it must do on 
    its next turn. If you issue no command, it spends its turn attacking any 
    creature within reach that has attacked it.
    At the end of each of the demon’s 
    turns, it makes a Charisma saving throw. The demon has disadvantage on this 
    saving throw if you say its true name. On a failed save, the demon continues to 
    obey you. On a successful save, your control of the demon ends for the rest of 
    the duration, and the demon spends its turns pursuing and attacking the nearest 
    non-demons to the best of its ability. If you stop concentrating on the spell 
    before it reaches its full duration, an uncontrolled demon doesn’t disappear for
     1d6 rounds if it still has hit points.
    As part of casting the spell, you can 
    form a circle on the ground with the blood used as a material component. The 
    circle is large enough to encompass your space. While the spell lasts, the 
    summoned demon can’t cross the circle or harm it, and it can’t target anyone 
    within it. Using the material component in this manner consumes it when the 
    spell ends.
    
    At Higher Levels: When you cast this spell using a spell slot of 
    5th level or higher, the challenge rating increases by 1 for each slot level 
    above 4th.
    """
    name = "Summon Greater Demon"
    level = 4
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A vial of blood from a humanoid killed within the past 24 hours"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Warlock', 'Wizard')


class SummonLesserDemons(Spell):
    """You utter foul words, summoning demons from the chaos of the Abyss. Roll on the 
    following table to determine what appears.
    d6 / Demons Summoned
    1–2 / Two demons
     of challenge rating 1 or lower
    3–4 / Four demons of challenge rating 1/2 or 
    lower
    5–6 / Eight demons of challenge rating 1/4 or lower
    
    The DM chooses the 
    demons, such as manes or dretches, and you choose the unoccupied spaces you can 
    see within range where they appear. A summoned demon disappears when it drops to
     0 hit points or when the spell ends.
    The demons are hostile to all creatures, 
    including you. Roll initiative for the summoned demons as a group, which has its
     own turns. The demons pursue and attack the nearest non-demons to the best of 
    their ability.
    As part of casting the spell, you can form a circle on the ground
     with the blood used as a material component. The circle is large enough to 
    encompass your space. While the spell lasts, the summoned demons can’t cross the
     circle or harm it, and they can’t target anyone within it. Using the material 
    component in this manner consumes it when the spell ends.
    
    At Higher Levels: 
    When you cast this spell using a spell slot of 6th or 7th level, you summon 
    twice as many demons. If you cast it using a spell slot of 8th or 9th level, you
     summon three times as many demons.
    """
    name = "Summon Lesser Demons"
    level = 3
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ('V', 'S', 'M')
    materials = """A vial of blood from a humanoid killed within the past 24 hours"""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Warlock', 'Wizard')


class Sunbeam(Spell):
    """A beam of brilliant light flashes out from your hand in a 5-foot-wide, 60-foot-
    line.
    Each creature in the line must make a Constitution saving throw. On a 
    failed save, a creature takes 6d8 radiant damage and is blinded until your next 
    turn. On a successful save, it takes half as much damage and isn’t blinded by 
    this spell. Undead and oozes have disadvantage on this saving throw.
    
    You can 
    create a new line of radiance as your action on any turn until the spell ends.
    
    
    For the duration, a mote of brilliant radiance shines in your hand. It sheds 
    bright light in a 30-foot radius and dim light for an additional 30 feet. The 
    light is sunlight.
    """
    name = "Sunbeam"
    level = 6
    casting_time = "1 action"
    casting_range = "Self (60-foot line)"
    components = ('V', 'S', 'M')
    materials = """A magnifying glass"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class Sunburst(Spell):
    """Brilliant sunlight flashes in a 60-foot radius centered on a point you choose 
    within range.
    Each creature in that light must make a Constitution saving throw.
     On a failed save, a creature takes 12d6 radiant damage and is blinded for 1 
    minute. On a successful save, it takes half as much damage and isn’t blinded by 
    this spell. Undead and oozes have disadvantage on this saving throw.
    
    A creature
     blinded by this spell makes another Constitution saving throw at the end of 
    each of its turns. On a successful save, it is no longer blinded.
    
    This spell 
    dispels any darkness in its area that was created by a spell.
    """
    name = "Sunburst"
    level = 8
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ('V', 'S', 'M')
    materials = """Fire and a piece of sunstone"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ('Druid', 'Sorcerer', 'Wizard')


class SwiftQuiver(Spell):
    """You transmute your quiver so it produces an endless supply of nonmagical 
    ammunition, which seems to leap into your hand when you reach for it.
    
    On each 
    of your turns until the spell ends, you can use a bonus action to make two 
    attacks with a weapon that uses ammunition from the quiver. Each time you make 
    such a ranged attack, your quiver magically replaces the piece of ammunition you
     used with a similar piece of nonmagical ammunition. Any pieces of ammunition 
    created by this spell disintegrate when the spell ends. If the quiver leaves 
    your possession, the spell ends.
    """
    name = "Swift Quiver"
    level = 5
    casting_time = "1 bonus action"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """A quiver containing at least one piece of ammunition"""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ('Ranger',)


class SwordBurst(Spell):
    """You create a momentary circle of spectral blades that sweep around you.
    Each 
    creature within range, other than you, must succeed on a Dexterity saving throw 
    or take 1d6 force damage.
    
    At Higher Levels: This spell's damage increases by 
    1d6 when you reach 5th level (2d6), 11th level (3d6), and 17th level (4d6).
    """
    name = "Sword Burst"
    level = 0
    casting_time = "1 action"
    casting_range = "5 feet"
    components = ('V',)
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ('Sorcerer', 'Warlock', 'Wizard')


class Symbol(Spell):
    """When you cast this spell, you inscribe a harmful glyph either on a surface (such
     as a section of floor, a wall, or a table) or within an object that can be 
    closed to conceal the glyph (such as a book, a scroll, or a treasure chest).
    If 
    you choose a surface, the glyph can cover an area of the surface no larger than 
    10 feet in diameter. If you choose an object, that object must remain in its 
    place; if the object is moved more than 10 feet from where you cast this spell, 
    the glyph is broken, and the spell ends without being triggered.
    
    The glyph is 
    nearly invisible, requiring an Intelligence (Investigation) check against your 
    spell save DC to find it.
    
    You decide what triggers the glyph when you cast the 
    spell. For glyphs inscribed on a surface, the most typical triggers include 
    touching or stepping on the glyph, removing another object covering it, 
    approaching within a certain distance of it, or manipulating the object that 
    holds it. For glyphs inscribed within an object, the most common triggers are 
    opening the object, approaching within a certain distance of it, or seeing or 
    reading the glyph.
    
    You can further refine the trigger so the spell is activated
     only under certain circumstances or according to a creature's physical 
    characteristics (such as height or weight), or physical kind (for example, the 
    ward could be set to affect hags or shapechangers). You can also specify 
    creatures that don't trigger the glyph, such as those who say a certain 
    password.
    
    When you inscribe the glyph, choose one of the options below for ist 
    effect. Once triggered, the glyph glows, filling a 60-foot-radius sphere with 
    dim light for 10 minutes, after which time the spell ends. Each creature in the 
    sphere when the glyph activates is targeted by its effect, as is a creature that
     enters the sphere for the first time on a turn or ends its turn there.
    
    Death
    
    Each target must make a Constitution saving throw, taking 10d10 necrotic damage 
    on a failed save, or half as much damage on a successful save.
    
    Discord
    Each 
    target must make a Constitution saving throw. On a failed save, a target bickers
     and argues with other creatures for 1 minute. During this time, it is incapable
     of meaningful communication and has disadvantage on attack rolls and ability 
    checks.
    Fear
    Each target must make a Wisdom saving throw and becomes frightened 
    for 1 minute on a failed save. While frightened, the target drops whatever it is
     holding and must move at least 20 feet away from the glyph on each of ist 
    turns, if able. 
    
    Hopelessness
    Each target must make a Charisma saving throw. On
     a failed save, the target is overwhelmed with despair for 1 minute. During this
     time, it can't attack or target any creature with harmful abilities, spells, or
     other magical effects.
    
    Insanity
    Each target must make an Intelligence saving 
    throw. On a failed save, the target is driven insane for 1 minute. An insane 
    creature can't take actions, can't understand what other creatures say, can't 
    read, and speaks only in gibberish. The DM controls its movement, which is 
    erratic.
    
    Pain
    Each target must make a Constitution saving throw and becomes 
    incapacitated with excruciating pain for 1 minute on a failed save.
    
    Sleep
    Each 
    target must make a Wisdom saving throw and falls unconscious for 10 minutes on a
     failed save. A creature awakens if it takes damage or if someone uses an action
     to shake or slap it awake.
    
    Stunning
    Each target must make a Wisdom saving 
    throw and becomes stunned for 1 minute on a failed save.
    """
    name = "Symbol"
    level = 7
    casting_time = "1 minute"
    casting_range = "Touch"
    components = ('V', 'S', 'M')
    materials = """Mercury, phosphorus, and powdered diamond and opal with a total value of at least 1,000 gp, which the spell consumes"""
    duration = "Until dispelled or triggered"
    ritual = False
    magic_school = "Abjuration"
    classes = ('Bard', 'Cleric', 'Wizard')


class SynapticStatic(Spell):
    """You choose a point within range and cause psychic energy to explode there. Each 
    creature in a 20-foot-radius sphere centered on that point must make an 
    Intelligence saving throw. A creature with an Intelligence score of 2 or lower 
    can’t be affected by this spell. A target takes 8d6 psychic damage on a failed 
    save, or half as much damage on a successful one.
    After a failed save, a target 
    has muddled thoughts for 1 minute. During that time, it rolls a d6 and subtracts
     the number rolled from all its attack rolls and ability checks, as well as its 
    Constitution saving throws to maintain concentration. The target can make an 
    Intelligence saving throw at the end of each of its turns, ending the effect on 
    itself on a success.
    """
    name = "Synaptic Static"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ('V', 'S')
    materials = """"""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ('Bard', 'Sorcerer', 'Warlock', 'Wizard')





from .spells import Spell, create_spell
from .spells_a import *
from .spells_b import *
from .spells_c import *
from .spells_d import *
from .spells_e import *
from .spells_f import *
from .spells_g import *
from .spells_h import *
from .spells_i import *
from .spells_j import *
from .spells_k import *
from .spells_l import *
from .spells_m import *
from .spells_n import *
from .spells_o import *
from .spells_p import *
from .spells_q import *
from .spells_r import *
from .spells_s import *
from .spells_t import *
from .spells_u import *
from .spells_v import *
from .spells_w import *
from .spells_x import *
from .spells_y import *
from .spells_z import *



from .features import Feature, FeatureSelector
from .. import (weapons, armor)


# PHB
class Rage(Feature):
    """In battle, you fight with primal ferocity. On your turn, you can enter a
    rage as a bonus action. While raging, you gain the following benefits if
    you aren’t wearing heavy armor:

    --You have advantage on Strength checks and
    Strength saving throws.

    --When you make a melee weapon attack using
    Strength, you gain a bonus to the damage roll that increases as you gain
    levels as a barbarian, as shown in the Rage Damage column of the Barbarian
    table.

    --You have resistance to bludgeoning, piercing, and slashing
    damage.

    If you are able to cast spells, you can’t cast them or concentrate on them
    while raging. Your rage lasts for 1 minute. It ends early if you are
    knocked unconscious or if your turn ends and you haven’t attacked a hostile
    creature since your last turn or taken damage since then. You can also end
    your rage on your turn as a bonus action. Once you have raged the number of
    times shown for your barbarian level in the Rages column of the Barbarian
    table, you must finish a long rest before you can rage again

    """
    _name = "Rage"
    source = "Barbarian"

    @property
    def name(self):
        level = self.owner.Barbarian.level
        num = 2
        if level >= 3:
            num = 3
        if level >= 6:
            num = 4
        if level >= 12:
            num = 5
        if level >= 17:
            num = 6
        if level >= 20:
            num = 100
        damage = '+2'
        if level >= 9:
            damage = "+3"
        if level >= 16:
            damage = "+4"
        return self._name + " ({:s}, {:d}x/LR)".format(damage, num)


class UnarmoredDefenseBarbarian(Feature):
    """While you are not wearing any armor, your Armor Class equals 10 + your
    Dexterity modifier + your Constitution modifier. You can use a shield and
    still gain this benefit.

    This bonus is computed in the AC given on the Character Sheet above.

    """
    name = "Unarmored Defense"
    source = 'Barbarian'


class RecklessAttack(Feature):
    """Starting at 2nd level, you can throw aside all concern for defense to
    attack with fierce desperation. When you make your first attack on your
    turn, you can decide to attack recklessly. Doing so gives you advantage on
    melee weapon attack rolls using Strength during this turn, but attack rolls
    against you have advantage until your next turn

    """
    name = "Reckless Attack"
    source = "Barbarian"


class DangerSense(Feature):
    """At 2nd level, you gain an uncanny sense of when things nearby aren’t as
    they should be, giving you an edge when you dodge away from danger. You
    have advantage on Dexterity saving throws against effects that you can see,
    such as traps and spells. To gain this benefit, you can’t be blinded,
    deafened, or incapacitated.

    """
    name = "Danger Sense"
    source = "Barbarian"


class ExtraAttackBarbarian(Feature):
    """Beginning at 5th level, you can attack twice, instead of once, whenever you
    take the Attack action on your turn.

    """
    name = "Extra Attack (2x)"
    source = "Barbarian"


class FastMovement(Feature):
    """Starting at 5th level, your speed increases by 10 feet while you aren’t
    wearing heavy armor.

    """
    name = "Fast Movement"
    source = "Barbarian"


class FeralInstinct(Feature):
    """By 7th level, your instincts are so honed that you have advantage on
    initiative rolls. Additionally, if you are surprised at the beginning of
    combat and aren’t incapacitated, you can act normally on your first turn,
    but only if you enter your rage before doing anything else on that turn.

    """
    name = "Feral Instinct"
    source = "Barbarian"


class BrutalCritical(Feature):
    """Beginning at 9th level, you can roll one additional weapon damage die when
    determining the extra damage for a critical hit with a melee attack. This
    increases to two additional dice at 13th level and three additional dice at
    17th level.

    """
    name = "Brutal Critical"
    source = "Barbarian"


class RelentlessRage(Feature):
    """Starting at 11th level, your rage can keep you fighting despite grievous
    wounds. If you drop to 0 hit points while you’re raging and don’t die
    outright, you can make a DC 10 Constitution saving throw. If you succeed,
    you drop to 1 hit point instead. Each time you use this feature after the
    first, the DC increases by 5. When you finish a short or long rest, the DC
    resets to 10.

    """
    name = "Relentless Rage"
    source = "Barbarian"


class PersistentRage(Feature):
    """Beginning at 15th level, your rage is so fierce that it ends early only if
    you fall unconscious or if you choose to end it.

    """
    name = "Persistent Rage"
    source = "Barbarian"

    
class IndomitableMight(Feature):
    """Beginning at 18th level, if your total for a Strength check is less than
    your Strength score, you can use that score in place of the total.

    """
    name = "Indomitable Might"
    source = "Barbarian"

    
class PrimalChampion(Feature):
    """At 20th level, you embody the power of the wilds. Your Strength and
    Constitution scores increase by 4. Your maximum for those scores is now 24.

    """
    name = "Primal Champion"
    source = "Barbarian"


# Berserker
class Frenzy(Feature):
    """Starting when you choose this path at 3rd level, you can go into a frenzy
    when you rage. If you do so, for the duration of your rage you can make a
    single melee weapon attack as a bonus action on each of your turns after
    this one. When your rage ends, you suffer one level o f exhaustion (as
    described in appendix A)

    """
    name = "Frenzy"
    source = "Barbarian (Berserker)"


class MindlessRage(Feature):
    """Beginning at 6th level, you can’t be charmed or frightened while raging. If
    you are charmed or frightened when you enter your rage, the effect is
    suspended for the duration of the rage.

    """
    name = "Mindless Rage"
    source = "Barbarian (Berserker)"


class IntimidatingPresence(Feature):
    """Beginning at 10th level, you can use your action to frighten someone with
    your menacing presence. When you do so, choose one creature that you can
    see within 30 feet of you. If the creature can see or hear you, it must
    succeed on a Wisdom saving throw (DC equal to 8 + your proficiency bonus +
    your Charisma modifier) or be frightened of you until the end of your next
    turn. On subsequent turns, you can use your action to extend the duration
    of this effect on the frightened creature until the end of your next
    turn. This effect ends if the creature ends its turn out of line of sight
    or more than 60 feet away from you. If the creature succeeds on its saving
    throw, you can't use this feature on that creature again for 24 hours.

    """
    name = "Intimidating Presence"
    source = "Barbarian (Berserker)"


class Retaliation(Feature):
    """Starting at 14th level, when you take damage from a creature that is within
    5 feet of you. you can use your reaction to make a melee weapon attack
    against that creature.

    """
    name = "Retaliation"
    source = "Barbarian (Berserker)"


# Totem Warrior
class SpiritSeeker(Feature):
    """Yours is a path that seeks attunement with the natural world, giving you a
    kinship with beasts. At 3rd level when you adopt this path, you gain the
    ability to cast the beast sense and speak with animals spells, but only as
    rituals, as described in chapter 10 of PHB.

    """
    name = "Spirit Seeker"
    source = "Barbarian (Totem Warrior)"

    
class BearSpirit(Feature):
    """While raging, you have resistance to all damage except psychic damage. The
    spirit of the bear makes you tough enough to stand up to any punishment.

    """
    name = "Totem Spirit (Bear)"
    source = "Barbarian (Totem Warrior)"


class EagleSpirit(Feature):
    """While you're raging and aren’t wearing heavy armor, other creatures have
    disadvantage on opportunity attack rolls against you, and you can use the
    Dash action as a bonus action on your turn. The spirit of the eagle makes
    you into a predator who can weave through the fray with ease.

    """
    name = "Totem Spirit (Eagle)"
    source = "Barbarian (Totem Warrior)"


class WolfSpirit(Feature):
    """While you're raging, your friends have advantage on melee attack rolls
    against any creature within 5 feet of you that is hostile to you. The
    spirit of the wolf makes you a leader of hunters

    """
    name = "Totem Spirit (Wolf)"
    source = "Barbarian (Totem Warrior)"


class ElkSpirit(Feature):
    """While you're raging and aren't wearing heavy armor, your walking speed
    increases by 15 feet. The spirit of the elk makes you extraordinarily swift

    """
    name = "Totem Spirit (Elk)"
    source = "Barbarian (Totem Warrior)"


class TigerSpirit(Feature):
    """While raging, you can add 10 feet to your long jump distance and 3 feet to
    your high jump distance. The spirit of the tiger empowers your leaps

    """
    name = "Totem Spirit (Tiger)"
    source = "Barbarian (Totem Warrior)"


class TotemSpirit(FeatureSelector):
    """
    Select a Totem Spirit from one of the following under feature_choices in
    your .py file:

    bear spirit

    eagle spirit

    wolf spirit

    elk spirit

    tiger spirit

    """
    options = {'bear spirit': BearSpirit,
               'eagle spirit': EagleSpirit,
               'wolf spirit': WolfSpirit,
               'elk spirit': ElkSpirit,
               'tiger spirit': TigerSpirit}
    name = "Totem Spirit (Select One)"
    source = "Barbarian (Totem Warrior)"


class BearAspect(FeatureSelector):
    """You gain the might of a bear. Your carrying capacity (including maximum
    load and maximum lift) is doubled, and you have advantage on Strength
    checks made to push, pull, lift, or break objects.

    """
    name = "Aspect of the Beast (Bear)"
    source = "Barbarian (Totem Warrior)"


class EagleAspect(FeatureSelector):
    """You gain the eyesight of an eagle. You can see up to 1 mile away with no
    difficulty, able to discern even fine details as though looking at
    something no more than 100 feet away from you. Additionally, dim light
    doesn't impose disadvantage on your Wisdom (Perception) checks.

    """
    name = "Aspect of the Beast (Eagle)"
    source = "Barbarian (Totem Warrior)"


class WolfAspect(FeatureSelector):
    """You gain the hunting sensibilities of a wolf. You can track other creatures
    while traveling at a fast pace, and you can move stealthily while traveling
    at a normal pace (see chapter 8 for rules on travel pace).

    """
    name = "Aspect of the Beast (Wolf)"
    source = "Barbarian (Totem Warrior)"


class ElkAspect(FeatureSelector):
    """Whether mounted or on foot , your travel pace is doubled, as is the travel
    pace of up to ten companions while they're within 60 feet of you and you're
    not incapacitated (see chapter 8 in the Player's Handbook for more
    information about travel pace). The elk spirit helps you roam far and fast

    """
    name = "Aspect of the Beast (Elk)"
    source = "Barbarian (Totem Warrior)"


class TigerAspect(FeatureSelector):
    """You gain proficiency in two skills from the following list: Athletics,
    Acrobatics, Stealth, and Survival. The cat spirit hones your survival
    instincts

    """
    name = "Aspect of the Beast (Tiger)"
    source = "Barbarian (Totem Warrior)"


class BeastAspect(FeatureSelector):
    """Select an Aspect of the Beast from one of the following under
    feature_choices in your .py file:

    bear aspect

    eagle aspect

    wolf aspect

    elk aspect

    tiger aspect

    """
    options = {'bear aspect': BearAspect,
               'eagle aspect': EagleAspect,
               'wolf aspect': WolfAspect,
               'elk aspect': ElkAspect,
               'tiger aspect': TigerAspect}
    name = "Aspect of the Beast (Select One)"
    source = "Barbarian (Totem Warrior)"


class SpiritWalker(Feature):
    """At 10th level, you can cast the commune with nature spell, but only as a
    ritual. When you do so, a spiritual version of one of the animals you chose
    for Totem Spirit or Aspect of the Beast appears to you to convey the
    information you seek.

    """
    name = "Spirit Walker"
    source = "Barbarian (Totem Warrior)"


class BearAttunement(Feature):
    """While you’re raging, any creature within 5 feet o f you that’s hostile to
    you has disadvantage on attack rolls against targets other than you or
    another character with this feature. An enemy is immune to this effect if
    it can’t see or hear you or if it can’t be frightened.
    
    """
    name = "Totemic Attunement (Bear)"
    source = "Barbarian (Totem Warrior)"


class EagleAttunement(Feature):
    """While raging, you have a flying speed equal to your current walking
    speed. This benefit works only in short bursts; you fall if you end your
    turn in the air and nothing else is holding you aloft.

    """
    name = "Totemic Attunement (Eagle)"
    source = "Barbarian (Totem Warrior)"


class WolfAttunement(Feature):
    """Wolf. While you’re raging, you can use a bonus action on your turn to knock
    a Large or smaller creature prone when you hit it with melee weapon attack.

    """
    name = "Totemic Attunement (Wolf)"
    source = "Barbarian (Totem Warrior)"


class ElkAttunement(Feature):
    """While raging, you can use a bonus action during your move to pass through
    the space of a Large or smaller creature. That creature must succeed on a
    Strength saving throw (DC 8 + your Strength bonus + your proficiency
    bonus) or be knocked prone and take bludgeoning damage equal to 1d12 + your
    Strength modifier
    
    """
    name = "Totemic Attunement (Elk)"
    source = "Barbarian (Totem Warrior)"


class TigerAttunement(Feature):
    """While you're raging, if you move at least 20 feet in a straight line toward
    a Large or smaller target right before making a melee weapon attack against
    it, you can use a bonus action to make an additional melee weapon attack
    against it

    """
    name = "Totemic Attunement (Tiger)"
    source = "Barbarian (Totem Warrior)"


class TotemicAttunement(FeatureSelector):
    """Select a Totemic Attunement from one of the following under feature_choices
    in your .py file:

    bear attunement

    eagle attunement

    wolf attunement

    elk attunement

    tiger attunement

    """
    options = {'bear attunement': BearAttunement,
               'eagle attunement': EagleAttunement,
               'wolf attunement': WolfAttunement,
               'elk attunement': ElkAttunement,
               'tiger attunement': TigerAttunement}
    name = "Totemic Attunement (Select One)"
    source = "Barbarian (Totem Warrior)"
    

# Battlerager
class BattleragerArmor(Feature):
    """When you choose this path at 3rd level, you gain the ability to use spiked
    armor (see the "Spiked Armor" sidebar in SCAG) as a weapon. While you are
    wearing spiked armor and are raging, you can use a bonus action to make one
    melee weapon attack with your armor spikes against a target within 5 feet
    of you. If the attack hits , the spikes deal ld4 piercing damage. You use
    your Strength modifier for the attack and damage rolls. Additionally, when
    you use the Attack action to grapple a creature, the target takes 3
    piercing damage if your grapple check succeeds.

    """
    name = 'Battlerager Armor'
    source = "Barbarian (Battlerager)"


class RecklessAbandon(Feature):
    """Beginning at 6th level, when you use Reckless Attack while raging, you also
    gain temporary hit points equal to your Constitution modifier (minimum of
    1). They vanish if any of them are left when your rage ends .

    """
    name = "Reckless Abandon"
    source = "Barbarian (Battlerager)"


class BattleragerCharge(Feature):
    """Beginning at 10th level, you can take the Dash action as a bonus action
    while you are raging.

    """
    name = "Battlerager Charge"
    source = "Barbarian (Battlerager)"

    
class SpikedRetribution(Feature):
    """Starting at 14th level, when a creature within 5 feet of you hits you with
    a melee attack, the attacker takes 3 piercing damage if you are raging,
    aren't incapacitated, and are wearing spiked armor.

    """
    name = "Spiked Retribution"
    source = "Barbarian (Battlerager)"


# Ancestral Guardian
class AncestralProtectors(Feature):
    """Starting when you choose this path at 3rd level, spectral warriors appear
    when you enter your rage. While you’re raging, the first creature you hit
    with an attack on your turn becomes the target of the warriors, which
    hinder its attacks. Until the start ofyour next turn, that target has
    disadvantage on any attack roll that isn’t against you, and when the target
    hits a creature other than you with an attack, that creature has resistance
    to the damage dealt by the attack. The effect on the target ends early
    ifyour rage ends

    """
    name = "Ancestral Protectors"
    source = "Barbarian (Ancestral Guardian)"


class SpiritShield(Feature):
    """Beginning at 6th level, the guardian spirits that aid you can provide
    supernatural protection to those you defend. If you are raging and another
    creature you can see within 30 feet of you takes damage, you can use your
    reaction to reduce that damage by 2d6.

    When you reach certain levels in this class. you can reduce the damage by
    more: by 3d6 at 10th level and by 4d6 at 14th level.

    """
    _name = "Spirit Shield"
    source = "Barbarian (Ancestral Guardian)"

    @property
    def name(self):
        level = self.owner.Barbarian.level
        damage = " (2d6)"
        if level >= 10:
            damage = " (3d6)"
        if level >= 14:
            damage = " (4d6)"
        return self._name + damage


class ConsultTheSpirits(Feature):
    """At 10th level, you gain the ability to consult with your ancestral
    spirits. When you do so, you cast the augury or clairvoyance spell, without
    using a spell slot or material components. Rather than creating a spherical
    sensor, this use of clairvoyance invisibly summons one Of your ancestral
    spirits to the chosen location. Wisdom is your spellcasting ability for
    these spells. After you cast either spell in this way, you can’t use this
    feature again until you finish a short or long rest

    """
    name = "Consult the Spirits"
    source = "Barbarian (Ancestral Guardian)"


class VengefulAncestors(Feature):
    """At 14th level, your ancestral spirits grow powerful enough to
    retaliate. When you use your Spirit Shield to reduce the damage of an
    attack, the attacker takes an amount of force damage equal to the damage
    that your Spirit Shield prevents.

    """
    name = "Vengeful Ancestors"
    source = "Barbarian (Ancestral Guardian)"


# Storm Herald
class DesertAura(Feature):
    """Starting at 3rd level, you emanate a stormy, magical aura while you
    rage. The aura extends 10 feet from you in every direction, but not through
    total cover. Your aura has an effect that activates when you enter your
    rage, and you can activate the effect again on each of your turns as a
    bonus action.

    Choose desert, sea, or tundra. Your aura’s effect depends on that chosen
    environment, as detailed below. You can change your environment choice
    whenever you gain a level in this class. If your aura's effects require a
    saving throw, the DC equals 8 + your proficiency bonus + your Constitu-
    tion modifier.

    **Desert**: When this effect is activated, all other creatures in your
    aura take 2 fire damage each. The damage increases when you reach certain
    levels in this class, increasing to 3 at 5th level, 4 at 10th level, 5 at
    15th level, and 6 at 20th level.

    """
    name = "Storm Aura (Desert)"
    source = "Barbarian (Storm Herald)"


class SeaAura(Feature):
    """Starting at 3rd level, you emanate a stormy, magical aura while you
    rage. The aura extends 10 feet from you in every direction, but not through
    total cover. Your aura has an effect that activates when you enter your
    rage, and you can activate the effect again on each of your turns as a
    bonus action.

    Choose desert, sea, or tundra. Your aura’s effect depends on that chosen
    environment, as detailed below. You can change your environment choice
    whenever you gain a level in this class. If your aura's effects require a
    saving throw, the DC equals 8 + your proficiency bonus + your Constitu-
    tion modifier.

    **Sea**: When this effect is activated, you can choose one other creature
    you can see in your aura. The target must make a Dexterity saving
    throw. The target takes 1d6 lightning damage on a failed save, or half as
    much damage on a successful one. The damage increases when you reach
    certain levels in this class, increasing to 2d6 at 10th level, 3d6 at 15th
    level, and 4d6 at 20th level.

    """
    name = "Storm Aura (Sea)"
    source = "Barbarian (Storm Herald)"


class TundraAura(Feature):
    """Starting at 3rd level, you emanate a stormy, magical aura while you
    rage. The aura extends 10 feet from you in every direction, but not through
    total cover. Your aura has an effect that activates when you enter your
    rage, and you can activate the effect again on each of your turns as a
    bonus action.

    Choose desert, sea, or tundra. Your aura’s effect depends on that chosen
    environment, as detailed below. You can change your environment choice
    whenever you gain a level in this class. If your aura's effects require a
    saving throw, the DC equals 8 + your proficiency bonus + your Constitu-
    tion modifier.

    **Tundra**: When this effect is activated, each creature of your choice in
    your aura gains 2 temporary hit points, as icy spirits inure it to
    suffering. The temporary hit points increase when you reach certain levels
    in this class, increasing to 3 at 5th level, 4 at 10th level, 5 at 15th
    level, and 6 at 20th level.

    """
    name = "Storm Aura (Tundra)"
    source = "Barbarian (Storm Herald)"


class StormAura(FeatureSelector):
    """
    Select a Storm Aura from one of the following under feature_choices in
    your .py file:

    desert

    sea

    tundra
    """
    options = {'desert': DesertAura,
               'sea': SeaAura,
               'tundra': TundraAura}
    name = "Storm Aura (Select One)"
    source = "Barbarian (Storm Herald)"


class DesertSoul(Feature):
    """At 6th level, the storm grants you benefits even when your aura isn’t
    active. The benefits are based on the environment you chose for your
    Storm Aura.

    **Desert**: You gain resistance to fire damage, and you don’t suffer the
    effects of extreme heat, as described in the Dungeon Master’s
    Guide. Moreover, as an action, you can touch a flammable object that isn’t
    being worn or carried by anyone else and set it on fire

    """
    name = "Storm Soul (Desert)"
    source = "Barbarian (Storm Herald)"


class SeaSoul(Feature):
    """At 6th level, the storm grants you benefits even when your aura isn’t
    active. The benefits are based on the environment you chose for your
    Storm Aura.

    **Sea**: You gain resistance to lightning damage, and
    you can breathe underwater. You also gain a swimming speed of 30 feet.
    """
    name = "Storm Soul (Sea)"
    source = "Barbarian (Storm Herald)"


class TundraSoul(Feature):
    """At 6th level, the storm grants you benefits even when your aura isn’t
    active. The benefits are based on the environment you chose for your
    Storm Aura.

    **Tundra**: You gain resistance to cold damage, and you don’t suffer the
    effects of extreme cold, as described in the Dungeon Master’s
    Guide. Moreover, as an action, you can touch water and turn a 5-foot cube
    Of it into ice, which melts after 1 minute. This action fails if a creature
    is in the cube
    """
    name = "Storm Soul (Tundra)"
    source = "Barbarian (Storm Herald)"


class StormSoul(FeatureSelector):
    """Select a Storm Soul (same as Soul Aura) from one of the following under
    feature_choices in your .py file:

    desert

    sea

    tundra

    """
    options = {'desert': DesertSoul,
               'sea': SeaSoul,
               'tundra': TundraSoul}
    name = "Storm Soul (Select One)"
    source = "Barbarian (Storm Herald)"


class ShieldingStorm(Feature):
    """At 10th level, you learn to use your mastery of the storm to protect
    others. Each creature of your choice has the damage resistance you gained
    from the Storm Soul fea ture while the creature is in your Storm Aura.

    """
    name = "Shielding Storm"
    source = "Barbarian (Storm Herald)"


class RagingDesert(Feature):
    """At 14th level, the power of the storm you channel grows mightier, lashing
    out at your foes. The effect is based on the environment you chose for your
    Storm Aura

    **Desert**: Immediately after a creature in your aura hits you with an
    attack, you can use your reaction to force that creature to make a
    Dexterity saving throw. On a failed save, the creature takes fire damage
    equal to half your barbarian level.

    """
    name = "Raging Storm (Desert)"
    source = "Barbarian (Storm Herald)"


class RagingSea(Feature):
    """At 14th level, the power of the storm you channel grows mightier, lashing
    out at your foes. The effect is based on the environment you chose for your
    Storm Aura

    **Sea**: When you hit a creature in your aura with an attack, you can use
    your reaction to force that creature to make a Strength saving throw. On a
    failed save, the creature is knocked prone, as if struck by a wave.

    """
    name = "Raging Storm (Sea)"
    source = "Barbarian (Storm Herald)"


class RagingTundra(Feature):
    """At 14th level, the power of the storm you channel grows mightier, lashing
    out at your foes. The effect is based on the environment you chose for your
    Storm Aura

    **Tundra**: Whenever the effect of your Storm Aura is activated, you can
    choose one creature you can see in the aura. That creature must succeed on
    a Strength saving throw, or its speed is reduced to 0 until the start of
    your next turn, as magical frost covers it

    """
    name = "Raging Storm (Tundra)"
    source = "Barbarian (Storm Herald)"


class RagingStorm(FeatureSelector):
    """Select a Raging Storm (same as Soul Aura) from one of the following under
    feature_choices in your .py file:

    desert

    sea

    tundra

    """
    options = {'desert': RagingDesert,
               'sea': RagingSea,
               'tundra': RagingTundra}
    name = "Raging Storm (Select One)"
    source = "Barbarian (Storm Herald)"


# Zealot
class DivineFury(Feature):
    """Starting when you choose this path at 3rd level, you can channel divine
    fury into your weapon strikes. While you’re raging, the first creature you
    hit on each of your turns with a weapon attack takes extra damage equal to
    1d6 + half your barbarian level. The extra damage is necrotic or radiant;
    you choose the type of damage when you gain this feature.

    """
    _name = "Divine Fury"
    source = "Barbarian (Zealot)"

    @property
    def name(self):
        level = self.owner.Barbarian.level
        damage = " (1d6+{:d})".format(level//2)
        return self._name + damage


class WarriorOfTheGods(Feature):
    """At 3rd level, your soul is marked for endless battle. If a spell, such as
    raise dead, has the sole effect of restoring you to life (but not
    undeath), the caster doesn’t need material components to cast the spell
    on you

    """
    name = "Warrior of the Gods"
    source = "Barbarian (Zealot)"


class FanaticalFocus(Feature):
    """Starting at 6th level, the divine power that fuels your rage can protect
    you. If you fail a saving throw while you’re raging, you can reroll it, and
    you must use the new roll. You can use this ability only once per rage.
    """
    name = "Fanatical Focus"
    source = "Barbarian (Zealot)"


class ZealousPresence(Feature):
    """At 10th level, you learn to channel divine power to inspire zealotry in
    others. As a bonus action, you unleash a battle cry infused with divine
    energy. Up to ten other creatures of your choice within 60 feet ofyou that
    can hear you gain advantage on attack rolls and saving throws until the
    start of your next turn. Once you use this feature, you can’t use it again
    until you finish a long rest

    """
    name = "Zealous Presence"
    source = "Barbarian (Zealot)"


class RageBeyondDeath(Feature):
    """Beginning at 14th level, the divine power that fuels your rage allows you
    to shrug off fatal blows. While you’re raging, having 0 hit points doesn’t
    knock you unconscious. You still must make death saving throws, and you
    suffer the normal effects of taking damage while at 0 hit points. However,
    if you would die due to failing death saving throws, you don't die until
    your rage ends, and you die then only if you still have 0 hit points.

    """
    name = "Rage Beyond Death"
    source = "Barbarian (Zealot)"
    



from collections import defaultdict
from ..features import Feature, FeatureSelector


class CharClass():
    """
    A generic Character Class (not to be confused with builtin class)
    """
    name = "Default"
    level = 1
    hit_dice_faces = 2
    subclass_select_level = 3
    weapon_proficiencies = ()
    _proficiencies_text = ()
    multiclass_weapon_proficiencies = ()
    _multiclass_proficiencies_text = ()
    saving_throw_proficiencies = ()
    primary_abilities = ()
    languages = ()
    class_skill_choices = ()
    num_skill_choices = 2
    spellcasting_ability = None
    spell_slots_by_level = None
    spells_known = ()
    spells_prepared = ()
    subclass = None
    subclasses_available = ()
    features_by_level = defaultdict(list)

    def __init__(self, level, owner=None, subclass=None, feature_choices=[],
                 **params):
        self.level = level
        self.owner = owner
        # For ex: add "char.Monk" attribute
        setattr(self.owner, self.name, self)
        # Instantiate the features
        self.features_by_level = defaultdict(list)
        cls = type(self)
        for i in range(1, 21):
            fs = []
            for f in cls.features_by_level[i]:
                if issubclass(f, FeatureSelector):
                    fs.append(f(owner=self.owner,
                                feature_choices=feature_choices))
                elif issubclass(f, Feature):
                    fs.append(f(owner=self.owner))
            self.features_by_level[i] = fs
        for k, v in params.items():
            setattr(self, k, v)
        self.spells_known = [S() for S in cls.spells_known]
        self.spells_prepared = [S() for S in cls.spells_prepared]

        # Apply subclass
        self.subclass = self.select_subclass(subclass)
        if isinstance(self.subclass, SubClass):
            self.apply_subclass(feature_choices=feature_choices)

    def select_subclass(self, subclass_str):
        """
        Return a SubClass object corresponding to given string.
        
        Intended to be replaced by classes so they can 
        define their own methods of picking subclass by string.
        """
        if subclass_str in ['', 'None', 'none', None]:
            return None
        for sc in self.subclasses_available:
            if subclass_str.lower() in sc.name.lower():
                return sc(owner=self.owner)
        return None

    def apply_subclass(self, feature_choices=[]):
        if not isinstance(self.subclass, SubClass):
            return
        subcls = self.subclass
        for i in range(1, 21):
            fs = []
            for f in subcls.features_by_level[i]:
                if issubclass(f, FeatureSelector):
                    fs.append(f(owner=self.owner,
                                feature_choices=feature_choices))
                elif issubclass(f, Feature):
                    fs.append(f(owner=self.owner))
            self.features_by_level[i].extend(fs)
        for attr in ('weapon_proficiencies', '_proficiencies_text'):
            new_list = tuple(getattr(self, attr, ())) + tuple(getattr(self.subclass, attr, ()))
            setattr(self, attr, new_list)
        # All subclass proficiencies transfer, regardless of if this is primary class
        self.multiclass_weapon_proficiencies += tuple(subcls.weapon_proficiencies)
        self._multiclass_proficiencies_text += tuple(subcls._proficiencies_text)
        self.spellcasting_ability = (self.spellcasting_ability or
                                     subcls.spellcasting_ability)
        self.spell_slots_by_level = (self.spell_slots_by_level or
                                     subcls.spell_slots_by_level)
        self.spells_known.extend([S() for S in subcls.spells_known])
        self.spells_prepared.extend([S() for S in subcls.spells_prepared])
    
    @property
    def features(self):
        features = ()
        for lvl in range(1, self.level+1):
            features += tuple(self.features_by_level[lvl])
        return features

    @property
    def is_spellcaster(self):
        result = (self.spellcasting_ability is not None)
        return result
    
    def spell_slots(self, spell_level):
        """How many spells slots are available for this spell level."""
        if self.spell_slots_by_level is None:
            return 0
        else:
            return self.spell_slots_by_level[self.level][spell_level]

    def __str__(self):
        s = 'Level {:d} {:s}'.format(self.level, self.name)
        if isinstance(self.subclass, SubClass):
            s += ' ({:s})'.format(str(self.subclass))
        return s
        
    def __repr__(self):
        return '\"{:s}\"'.format(str(self))


class SubClass():
    """
    A generic subclass object. Add more detail in the __doc__ attribute.
    """
    name = ''
    features_by_level = defaultdict(list)
    weapon_proficiencies = ()
    _proficiencies_text = ()
    spellcasting_ability = None
    spell_slots_by_level = None
    spells_known = ()
    spells_prepared = ()

    def __init__(self, owner):
        self.owner = owner
        self.__doc__ = self.__doc__ or SubClass.__doc__

    def __str__(self):
        return self.name

    def __repr__(self):
        return "\"{:s}\"".format(self.name)



#!/usr/bin/env python

import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='dungeonsheets',
      version=read('VERSION'),
      description='Dungeons and Dragons 5e Character Tools',
      long_description=read('README.rst'),
      long_description_content_type='text/x-rst',
      keywords='D&D character sheets',
      author='Mark Wolfman',
      author_email='canismarko@gmail.com',
      license='GPLv3',
      url='https://github.com/canismarko/dungeon-sheets',
      download_url = 'https://github.com/canismarko/dungeon-sheets/archive/master.zip',
      packages=find_packages(),
      package_data={
          'dungeonsheets': ['forms/*pdf', 'forms/*.tex', 'forms/*.txt',
                            '../VERSION']
      },
      install_requires=[
          'fdfgen', 'npyscreen', 'jinja2', 'pdfrw',
      ],
      entry_points={
          'console_scripts': [
              'makesheets = dungeonsheets.make_sheets:main',
              'create-character = dungeonsheets.create_character:main',
          ]
      },
      python_requires='>=3.6',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Games/Entertainment :: Role-Playing',
      ],
     )
